/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#pragma once

namespace oe {

//! A template COM smart pointer.
/*! A template COM smart pointer. This template class does not override 
    <code>operator\&</code>, so is safe to use in stl containers.
*/
template<typename I>
class ComPtr
{
private:
	I* p;
    
	void Reset(I* const& p_) throw();
	void Free() throw();
    
public:
    //! Creates an empty smart pointer.
	ComPtr() throw();
    
    //! Releases the encapsulated object.
    /*! Releases the encapsulated object, if present, using \c Release.
    */
    ~ComPtr() throw();
	
    //! Creates a smart pointer and assigns an object to it. 
    /*! Creates a smart pointer and assigns an object to it. \c AddRef is not called
        on the contents of \a p_. If \a p_ is 0, then effectively an empty smart pointer
        is created.
        \param p_   The object to be assigned, or 0 to create an empty smart pointer.
    */
	explicit ComPtr(I* const& p_) throw();
    
    //! Creates a copy of an existing smart pointer.
	ComPtr(const ComPtr<I>& ptr) throw();
    
    //! Copies an existing smart pointer.
	ComPtr<I>& operator=(const ComPtr<I>& ptr) throw();

	//! Returns the contents of the smart pointer.
	/*! Returns the contents of the smart pointer. \c AddRef is not called on the
        contents before this method returns.
        \return     The contents of the smart pointer, or 0 if it is empty.
    */
    I* const& Get() const throw();
	
    //! Returns a pointer-to-member of the smart pointer contents.
    /*! Returns a pointer-to-member of the smart pointer contents. This allows the smart pointer to be
        used as if it was a simple C++ pointer of type \c I*.
        \return     The contents of the smart pointer.
        \exception  Exception is thrown if the smart pointer is empty.
    */
    I* const& operator->() const throw(Exception);

	//! Returns \c true if the wrapper contains a pointer, \c false if empty.
	operator bool() const throw();

	//! Returns the address of the encapsulated pointer.
    /*! Returns the address of the encapsulated pointer. This is available to allow
        smart pointers to be used as targets for functions such as \c QueryInterface
        and \c CoCreateObject.
        \return     The address of the encapsulated pointer.
        \warning    If the address of a non-empty smart pointer is taken, a warning 
        is logged to \c cerr, since such usage can cause memory leaks.
    */
	I** AddressOf() throw();
};

//! Returns the standard comparison in pointer arithmetic.
template<typename I>
bool operator<(const ComPtr<I>& l, const ComPtr<I>& r) throw();

} // namespace oe

// Template implementation
#include "ComPtr.inl"
