/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"

namespace oe {

void GetBufferTraits(const OE_BUFFERTRAITS& traits, ulong& usageTraits, ulong& lockTraits) throw(Exception)
{
	switch(traits)
	{
	case OEBT_DYNAMIC:
		usageTraits = D3DUSAGE_DYNAMIC | D3DUSAGE_WRITEONLY;
		lockTraits = D3DLOCK_DISCARD;
		break;

	case OEBT_STATIC:
		usageTraits = D3DUSAGE_WRITEONLY;
		lockTraits = 0;
		break;

	default:
		throw Exception("Graphics::GetBufferTraits(): Unknown traits type.");
	}
}

Graphics::Graphics(const HINSTANCE& instance_) throw(Exception)
  : instance(instance_),
  	direct3d(Direct3DCreate8(D3D_SDK_VERSION)), 
	depthStencilFormat(D3DFMT_UNKNOWN)
{
	// create a Direct3D8 object (writes without AddRef)
	if(!direct3d)
		throw Exception("Graphics::Graphics(): Direct3DCreate8 failed - unable to create a Direct3D8 device."
			"\nPerhaps your installation of DirectX needs upgrading to version 8.1?");
}

void Graphics::SetDepthStencilFormat(const D3DFORMAT& format) throw()
{
	depthStencilFormat = format;
}

/*! \todo Make the number of back buffers application-defined.
	\todo Allow the device to be created with the D3DCREATE_FPU_PRESERVE flag set.
*/
bool Graphics::CreateFromLoader(const char* const& title) throw(Exception)
{
	long adapter = 0;
	D3DPRESENT_PARAMETERS pp;
	memset(&pp, 0, sizeof(D3DPRESENT_PARAMETERS));

	// show the loader and fill in the relevant presentation parameters
	GraphicsLoader loader(instance, direct3d, depthStencilFormat);
	if(!loader.Show(adapter, pp))
		return false;

	// log the device choice
	D3DADAPTER_IDENTIFIER8 ident;
	HRESULT hr = direct3d->GetAdapterIdentifier(adapter, D3DENUM_NO_WHQL_LEVEL, &ident);
    if(FAILED(hr))
		throw CreateDirectXException("Graphics::CreateFromLoader(): GetAdapterIdentifier failed: ", hr);
	std::cout << "Creating device: " << ident.Description << " (" << ident.Driver << "), ";

	// log the mode choice
	std::cout << "mode: " << pp.BackBufferWidth << "x" << pp.BackBufferHeight << " ";
	std::cout << (pp.Windowed ? "windowed" : "fullscreen");
	std::cout << "..." << std::endl;
	
	// fill in the rest of the parameters
	pp.BackBufferCount = 2;
	pp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	
	// compute the aspect ratio and create the window
	targetWidth = pp.BackBufferWidth;
	targetHeight = pp.BackBufferHeight;
	targetAspectRatio = _R(4) / _R(3);
	if(pp.Windowed)
	{
		D3DDISPLAYMODE current;
		HRESULT hr = direct3d->GetAdapterDisplayMode(adapter, &current);
		if(FAILED(hr))
			throw CreateDirectXException("Graphics::CreateFromLoader(): GetAdapterDisplayMode failed: ", hr);
		targetAspectRatio *= _R(targetWidth * current.Height) / _R(targetHeight * current.Width);
		window.Create(instance, title, pp.BackBufferWidth, pp.BackBufferHeight);
	}
	else
		window.Create(instance, title, 200, 100);
	
	// create the device
	hr = direct3d->CreateDevice(
		adapter, 
		D3DDEVTYPE_HAL, 
		window.Get(), 
		D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_PUREDEVICE, 
		&pp, 
		device.AddressOf()
	);
	if(FAILED(hr))
		throw CreateDirectXException("Graphics::CreateFromLoader(): CreateDevice failed: ", hr);

	// created!
	return true;
}

IDirect3DDevice8* const& Graphics::Get() const throw()
{ 
	return device.Get();
}

const HWND& Graphics::GetWindow() const throw()
{
	return window.Get();
}

IDirect3DDevice8* const& Graphics::operator->() const throw(Exception) 
{ 
	return device.Get(); 
}

const long& Graphics::GetTargetWidth() const throw() 
{ 
	return targetWidth; 
}

const long& Graphics::GetTargetHeight() const throw() 
{ 
	return targetHeight; 
}

const real_t& Graphics::GetTargetAspectRatio() const throw() 
{ 
	return targetAspectRatio; 
}

FrameTimer& Graphics::GetFrameTimer() throw()
{
	return timer;
}

bool Graphics::ProcessThreadMessages() throw()
{
	MSG msg;
	bool paused = false;
	for(;;)
	{
		// purge ye olde message queue
		while(PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
		{
			// quit on WM_QUIT
			if(msg.message == WM_QUIT)
				return true;
			// otherwise process normally
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		// only return if active
		if(RenderWindow::IsActive())
		{
			if(paused || RenderWindow::WasSizeMoved()) 
				timer.IgnoreThisFrame();
			return false;
		}
		// otherwise wait until we're active
		paused = true;
		WaitMessage();
	}
}

VertexShader Graphics::CreateVertexShader(ulong* const& decl, ulong* const& body) const throw(Exception)
{
	// log creation
	std::cout << "Creating vertex shader..." << std::endl;
	// create the shader handle
	ulong handle = 0;
	HRESULT hr = device->CreateVertexShader(decl, body, &handle, 0);
	if(FAILED(hr))
		throw CreateDirectXException("Graphics::CreateVertexShader(): CreateVertexShader failed: ", hr);
	// wrap the handle as a shared resource
	return VertexShader(handle, device);
}
	
PixelShader Graphics::CreatePixelShader(ulong* const& function) const throw(Exception)
{
	// log creation
	std::cout << "Creating pixel shader..." << std::endl;
	// create the shader handle
	ulong handle;
	HRESULT hr = device->CreatePixelShader(function, &handle);
	if(FAILED(hr))
		throw CreateDirectXException("Graphics::CreatePixelShader(): CreatePixelShader failed: ", hr);
	// wrap the handle as a shared resource
	return PixelShader(handle, device);
}


Texture Graphics::LoadTexture(const char* const& filename, const D3DFORMAT& format, const D3DPOOL& pool, const bool& mipmap) const throw(Exception)
{
	// log creation
	std::cout << "Loading texture \"" << filename << "\"..." << std::endl;

	// load it
	Texture texture;
	HRESULT hr = D3DXCreateTextureFromFileEx(
		device.Get(), 
		filename, 
		D3DX_DEFAULT,				// width
		D3DX_DEFAULT,				// height
		mipmap ? D3DX_DEFAULT : 1, 
		0,							// usage
		format, 
		pool, 
		D3DX_DEFAULT,				// filter
		D3DX_DEFAULT,				// mipfilter
		0,							// colour key
		0,							// src info
		0,							// palette
		texture.AddressOf()
	);
	if(FAILED(hr))
		throw CreateDirectXException("Graphics::LoadTexture(): D3DXCreateTextureFromFileEx failed: ", hr);
    // return a copy
	return texture;
}

Font Graphics::LoadFont(const char* const& fontname) const throw(Exception)
{
	// log creation
	std::cout << "Loading font \"" << fontname << "\"..." << std::endl;

	// create a texture
	std::string filename;
	filename.assign(fontname).append(".tga");
	Texture texture = LoadTexture(filename.c_str(), D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, false);
	
	// load up the character widths
	boost::shared_array<uchar> widths(new uchar[256]);
	filename.assign(fontname).append(".w");
	std::ifstream w(filename.c_str(), std::ios::binary);
	if(w.read(reinterpret_cast<char*>(widths.get()), 256).fail())
		throw Exception("Graphics::LoadFont(): ifstream::read failed.");
	
	// get the texture dimensions for the font height
	D3DSURFACE_DESC desc;
	HRESULT hr = texture->GetLevelDesc(0, &desc);
	if(FAILED(hr))
		throw CreateDirectXException("Graphics::LoadFont(): GetLevelDesc failed: ", hr);
	uchar height = static_cast<uchar>(desc.Height / 16);
	
	// return the font
	return Font(texture, widths, height);
}

void Graphics::ReleaseStreamSource(const ulong& stream) const throw()
{
	if(device)
		device->SetStreamSource(stream, 0, 0);
}

void Graphics::ReleaseTexture(const ulong& stage) const throw()
{
	if(device)
		device->SetTexture(stage, 0);
}

void Graphics::SetVertexShader(const VertexShader& shader) const throw(Exception)
{
	if(!shader.Get())
		throw Exception("Graphics::SetVertexShader(): Null shader.");
	HRESULT hr = device->SetVertexShader(shader.Get());
	if(FAILED(hr))
		throw CreateDirectXException("Graphics::SetVertexShader(): SetVertexShader failed: ", hr);
}

void Graphics::SetPixelShader(const PixelShader& shader) const throw(Exception)
{
	if(!shader.Get())
		throw Exception("Graphics::SetPixelShader(): Null shader.");
	HRESULT hr = device->SetPixelShader(shader.Get());
	if(FAILED(hr))
		throw CreateDirectXException("Graphics::SetPixelShader(): SetPixelShader failed: ", hr);
}

const HINSTANCE& Graphics::GetInstance() const throw()
{
	return instance;
}

IndexBuffer<ushort> Graphics::CreateIndexBuffer(const ulong& numIndices, const OE_BUFFERTRAITS& traits) const throw(Exception)
{
	// log creation
	std::cout << "Creating index buffer (" << numIndices * sizeof(ushort) << " bytes)..." << std::endl;
	// get the traits 
	ulong usageTraits, lockTraits;
	GetBufferTraits(traits, usageTraits, lockTraits);
	// create the buffer
	IndexBuffer<ushort> buffer(lockTraits);
	HRESULT hr = device->CreateIndexBuffer(
        numIndices * sizeof(ushort), 
		usageTraits, 
		D3DFMT_INDEX16, 
		D3DPOOL_DEFAULT, 
		buffer.AddressOf()
	);
	if(FAILED(hr))
		throw CreateDirectXException("Graphics::CreateIndexBuffer(): CreateIndexBuffer failed: ", hr);
	// return a copy
	return buffer;
}

void Graphics::SetVertexShaderConstant(const ulong& reg, const Vector3& vector) const throw(Exception)
{
	const float value[] = { vector.x, vector.y, vector.z, 1.0f };
	HRESULT hr = device->SetVertexShaderConstant(reg, value, 1);
	if(FAILED(hr))
		throw CreateDirectXException("Graphics::SetVertexShaderConstant(): SetVertexShaderConstant failed: ", hr);
}

void Graphics::SetPixelShaderConstant(const ulong& reg, const Vector3& vector) const throw(Exception)
{
	const float value[] = { vector.x, vector.y, vector.z, 1.0f };
	HRESULT hr = device->SetPixelShaderConstant(reg, value, 1);
	if(FAILED(hr))
		throw CreateDirectXException("Graphics::SetPixelShaderConstant(): SetPixelShaderConstant failed: ", hr);
}

void Graphics::SetVertexShaderConstant(const ulong& reg, const Matrix4x4& matrix, const ulong& rows) const throw(Exception)
{
	HRESULT hr = device->SetVertexShaderConstant(reg, &matrix, rows);
	if(FAILED(hr))
		throw CreateDirectXException("Graphics::SetVertexShaderConstant(): SetVertexShaderConstant failed: ", hr);
}

} // namespace oe
