/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#pragma once

namespace oe {

//! Used to control buffer creation and locking semantics.
enum OE_BUFFERTRAITS
{
	OEBT_DYNAMIC,	//!< A write only dynamic buffer.
	OEBT_STATIC		//!< A write only static buffer
};

//! Gets the creation and usage traits for a particular buffer type.
void GetBufferTraits(const OE_BUFFERTRAITS& traits, ulong& usageTraits, ulong& lockTraits) throw(Exception);

//! A wrapper around a DirectX 8 Graphics device.
class Graphics : boost::noncopyable
{
private:
	HINSTANCE instance;
	ComPtr<IDirect3D8> direct3d;
	ComPtr<IDirect3DDevice8> device;
	RenderWindow window;
	FrameTimer timer;

	long targetWidth, targetHeight;
	real_t targetAspectRatio;
	D3DFORMAT depthStencilFormat;

public:
    //! Creates a new graphics object for the given application instance.
    /*! Creates a new graphics object for the given application instance. This function
        also attempts to acquire an IDirect3D8 object.
        \exception Exception is thrown if an IDirect3D8 object cannot be acquired.
        \exception Exception is thrown if the timer resolution is too coarse.
    */
	Graphics(const HINSTANCE& instance_) throw(Exception);

    //! Sets the desired depth stencil format for this adapter.
    /*! Sets the desired depth stencil format for this adapter. This must be called before
        CreateFromLoader(). Set a format of D3DFMT_UNKNOWN to create no depth buffer, which
        is the default.
        \param format   The desired format, which must be a valid depth stencil format.
    */
	void SetDepthStencilFormat(const D3DFORMAT& format) throw();
    
    //! Create a device using a loader dialog.
    /*! Create a device using a loader dialog. The loader dialog enumerates all modes and 
        devices on the system that support pixel shaders and the current depth stencil format.
        \param title    The title for the render target window.
        \return \c true if the user-selected device is created, false if the user cancelled the dialog.
        \exception Exception is thrown if the user-selected device cannot be created.
    */
	bool CreateFromLoader(const char* const& title) throw(Exception);

    //! Gets the underlying IDirect3DDevice8 object.
	IDirect3DDevice8* const& Get() const throw();

	//! Gets the underlying window handle.
	const HWND& GetWindow() const throw();

	//! Gets the underlying application instance.
	const HINSTANCE& GetInstance() const throw();
    
    //! Returns a pointer-to-member of the underlying IDirect3DDevice8 object.
	IDirect3DDevice8* const& operator->() const throw(Exception);

	const long& GetTargetWidth() const throw();  //!< Get the target window width.
	const long& GetTargetHeight() const throw(); //!< Get the target window height.
	
    //! Get the target window aspect ratio.
    /*! Get the target window aspect ratio, this is not necessarily the same as \f$\frac{width}{height}\f$
        since the actual monitor aspect ratio is also taken into account.
    */
    const real_t& GetTargetAspectRatio() const throw();
	
	//! Return a reference to the frame timer.
	FrameTimer& GetFrameTimer() throw();
	
	//! Process the thread message queue.
	bool ProcessThreadMessages() throw();

	//! Creates a new vertex shader.
	VertexShader CreateVertexShader(ulong* const& decl, ulong* const& body) const throw(Exception);
	
	//! Creates a new pixel shader.
	PixelShader CreatePixelShader(ulong* const& function) const throw(Exception);

	//! Creates a new vertex buffer.
	template<typename V>
	VertexBuffer<V> CreateVertexBuffer(const ulong& numVertices, const OE_BUFFERTRAITS& traits) const throw(Exception);

	//! Creates a new index buffer of 16-bit indices.
	IndexBuffer<ushort> CreateIndexBuffer(const ulong& numIndices, const OE_BUFFERTRAITS& traits) const throw(Exception);

	//! Sets a vertex buffer stream source.
	template<typename V>
	void SetStreamSource(const ulong& stream, const VertexBuffer<V>& buffer) const throw(Exception);

	//! Unsets a vertex buffer stream source.
	void ReleaseStreamSource(const ulong& stream) const throw();

	//! Creates a new texture from an image on disk.
	Texture LoadTexture(const char* const& filename, const D3DFORMAT& format, const D3DPOOL& pool, const bool& mipmap) const throw(Exception);

	//! Loads a font from a texture and widths file on disk.
	Font LoadFont(const char* const& fontname) const throw(Exception);

	//! Sets the texture for a given texture stage.
	template<typename I>
	void SetTexture(const ulong& stage, const ComPtr<I>& texture) const throw(Exception);

	//! Unsets the texture from a given texture stage.
	void ReleaseTexture(const ulong& stage) const throw();

	//! Set the current vertex shader.
	void SetVertexShader(const VertexShader& shader) const throw(Exception);
	
	//! Set the current pixel shader.
	void SetPixelShader(const PixelShader& shader) const throw(Exception);

	//! Loads a vector into a vertex shader constant register.
	void SetVertexShaderConstant(const ulong& reg, const Vector3& vector) const throw(Exception);

	//! Loads all or part of a matrix into vertex shader constant registers.
	void SetVertexShaderConstant(const ulong& reg, const Matrix4x4& matrix, const ulong& rows = 4) const throw(Exception);

	//! Loads a vector into a pixel shader constant register.
	void SetPixelShaderConstant(const ulong& reg, const Vector3& vector) const throw(Exception);
};

} // namespace oe

// Template implementation
#include "Graphics.inl"
