/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"

namespace oe {

GraphicsLoader* GraphicsLoader::Get(const HWND& dialog) throw(Exception)
{
	// get the loader from the long pointer associated with the window
	LONG_PTR temp = GetWindowLongPtr(dialog, GWLP_USERDATA);
	if(!temp)
		throw Exception("GraphicsLoader::Get(): GetWindowLongPtr returned 0 - internal error.");
	return reinterpret_cast<GraphicsLoader*>(temp);
}

INT_PTR CALLBACK GraphicsLoader::DialogProc(HWND dialog, UINT msg, WPARAM wParam, LPARAM lParam) throw()
{
	try
	{
		switch(msg)
		{
		case WM_INITDIALOG:
			{
				// get and store the calling instance
				if(!lParam)
					throw Exception("GraphicsLoader::DialogProc(): Invalid this pointer passed during WM_INITDIALOG - internal error.");
				GraphicsLoader* const this_ = reinterpret_cast<GraphicsLoader*>(lParam);
#pragma warning(disable: 4244)
				SetWindowLongPtr(dialog, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(this_));
#pragma warning(default: 4244)

				// show and update the window
				SendDlgItemMessage(dialog, IDC_WINDOWED, BM_SETCHECK, BST_CHECKED, 0);
				ShowWindow(dialog, SW_SHOW);
				UpdateWindow(dialog);

				// load the adapters
				this_->EnumerateAdapters(dialog);
				this_->EnumerateModes(dialog);
			}
			return TRUE;

		case WM_COMMAND:
			switch(wParam)
			{
			case ((CBN_SELCHANGE << 16) | IDC_ADAPTER):
			case ((BN_CLICKED << 16) | IDC_WINDOWED):
				{
					// get the loader
					GraphicsLoader* const this_ = Get(dialog);
					// update the adapter, windowedness and modes
					this_->adapter = static_cast<long>(SendDlgItemMessage(dialog, IDC_ADAPTER, CB_GETCURSEL, 0, 0));
					this_->fullscreen = (SendDlgItemMessage(dialog, IDC_WINDOWED, BM_GETCHECK, 0, 0) != BST_CHECKED);
					this_->EnumerateModes(dialog);
				}
				return TRUE;

			case ((LBN_SELCHANGE << 16) | IDC_MODES):
				{
					// get the loader
					GraphicsLoader* const this_ = Get(dialog);
					// update the mode
					this_->mode = static_cast<long>(SendDlgItemMessage(dialog, IDC_MODES, LB_GETITEMDATA, 
						SendDlgItemMessage(dialog, IDC_MODES, LB_GETCURSEL, 0, 0), 0));
				}
				return TRUE;
			
			case ((LBN_DBLCLK << 16) | IDC_MODES):
			case IDOK:
				// accept the dialog
				EndDialog(dialog, 0);
				return TRUE;

			case IDCANCEL:
				// cancel the dialog
				EndDialog(dialog, 0xdeadbeef);
				return TRUE;
			}
			break;

		case WM_CLOSE:
			// cancel the dialog
			EndDialog(dialog, 0xdeadbeef);
			return TRUE;
		}
	}
	catch(Exception& e)
	{
		ReportException(e);
		EndDialog(dialog, 0xdeadbeef);
		return TRUE;
	}
	catch(...)
	{
		ReportException(Exception("Unknown exception caught!"));
		EndDialog(dialog, 0xdeadbeef);
		return TRUE;
	}
	// default to the standard message processor
	return FALSE;
}

void GraphicsLoader::EnumerateAdapters(const HWND& dialog) throw(Exception)
{
	// empty the current list
	SendDlgItemMessage(dialog, IDC_ADAPTER, CB_RESETCONTENT, 0, 0);

	// get the number of adapters
	ulong numAdapters = direct3d->GetAdapterCount();
	if(!numAdapters)
		throw Exception("GraphicsLoader::EnumerateAdapters(): GetAdapterCount returned 0 adapters on this system.");
	
	// load in each adapter
	for(ulong i = 0; i < numAdapters; ++i)
	{
		D3DADAPTER_IDENTIFIER8 ident;
		HRESULT hr = direct3d->GetAdapterIdentifier(i, D3DENUM_NO_WHQL_LEVEL, &ident);
		if(FAILED(hr))
			throw CreateDirectXException("GraphicsLoader::EnumerateAdapters(): GetAdapterIdentifier failed: ", hr);

		SendDlgItemMessage(dialog, IDC_ADAPTER, CB_ADDSTRING, 0, reinterpret_cast<LPARAM>(ident.Description));
	}

	// select the current one
	SendDlgItemMessage(dialog, IDC_ADAPTER, CB_SETCURSEL, adapter, 0);
}

bool GraphicsLoader::mode_comp::operator()(DisplayMode* const& l, DisplayMode* const& r) const throw()
{
	return l->mode.Width * l->mode.Height > r->mode.Width * r->mode.Height;
}

void GraphicsLoader::EnumerateModes(const HWND& dialog) throw(Exception)
{
	// empty the current list
	SendDlgItemMessage(dialog, IDC_MODES, LB_RESETCONTENT, 0, 0);
	
	// get the device caps
	D3DCAPS8 caps;
	HRESULT hr = direct3d->GetDeviceCaps(adapter, D3DDEVTYPE_HAL, &caps);
	if(FAILED(hr))
		throw CreateDirectXException("GraphicsLoader::EnumerateModes(): GetDeviceCaps failed: ", hr);
	
	// check the vertex and pixel shader versions
    caps.VertexShaderVersion &= 0xffff;
    caps.PixelShaderVersion &= 0xffff; // the top 2 bytes usually contain garbage
	if(caps.VertexShaderVersion < 0x0100 || caps.PixelShaderVersion < 0x0101)
	{
		MessageBox(dialog, "This adapter fails the minimum vertex and pixel shader requirements"
			"\n(1.0 and 1.1 respectively). Please choose a different adapter.", 
			"Invalid Adapter", MB_OK | MB_ICONINFORMATION | MB_TASKMODAL
		);
		Enable(dialog, false);
		return;
	}

	// check for support of the chosen depth stencil format
	if(depthStencilFormat != D3DFMT_UNKNOWN)
	{
		hr = direct3d->CheckDeviceFormat(
			adapter, D3DDEVTYPE_HAL, D3DFMT_X8R8G8B8,
			D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE, depthStencilFormat
		);
		if(FAILED(hr))
		{
			std::ostringstream ss;
			ss << "This adapter doesn't support the required depth stencil format: "
				<< FormatDirectXError(hr) << "\nPlease choose a different adapter.";
			MessageBox(dialog, ss.str().c_str(), "Invalid Adapter", MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);
			Enable(dialog, false);
			return;
		}
	}

	// get the current display mode
	D3DDISPLAYMODE current;
	hr = direct3d->GetAdapterDisplayMode(adapter, &current);
	if(FAILED(hr))
		throw CreateDirectXException("GraphicsLoader::EnumerateModes(): GetAdapterDisplayMode failed: ", hr);
	
	// get all the modes this adapter supports
	ulong numModes = direct3d->GetAdapterModeCount(adapter);
	if(!numModes)
		throw Exception("GraphicsLoader::EnumerateModes(): GetAdapterModeCount returned 0 - internal error.");
	boost::scoped_array<DisplayMode> modes(new DisplayMode[numModes]);
	for(ulong i = 0; i < numModes; ++i)
	{
		hr = direct3d->EnumAdapterModes(adapter, i, &modes[i].mode);
		if(FAILED(hr))
			throw CreateDirectXException("GraphicsLoader::EnumerateModes(): EnumAdapterModes failed: ", hr);
		modes[i].index = i;
	}

    // filter all except the highest refresh for supported modes
	std::hash_map<ulong, DisplayMode*> topmodes;
	for(ulong i = 0; i < numModes; ++i)
	{
		if(modes[i].mode.Format == D3DFMT_X8R8G8B8 || modes[i].mode.Format == D3DFMT_A8R8G8B8)
		{
			ulong key = ((modes[i].mode.Width & 0xffff) << 16) | (modes[i].mode.Height & 0xffff);
			if(fullscreen)
			{
				std::hash_map<ulong, DisplayMode*>::const_iterator it = topmodes.find(key);
				if(it == topmodes.end() || modes[i].mode.RefreshRate > it->second->mode.RefreshRate)
					topmodes[key] = &modes[i];
			}
			else if(
				modes[i].mode.Width < current.Width && 
				modes[i].mode.Height < current.Height && 
				modes[i].mode.RefreshRate == current.RefreshRate
			)
				topmodes[key] = &modes[i];
		}
	}

	// sanity check - we have at least 1 mode right?
	if(topmodes.empty())
	{
		MessageBox(dialog, "This adapter does not support a 32-bit frame buffer. Please choose a different adapter.", 
			"Invalid Adapter", MB_OK | MB_ICONINFORMATION | MB_TASKMODAL
		);
		Enable(dialog, false);
		return;
	}

	// sort the list
	std::vector<DisplayMode*> sortedmodes;
	for(std::hash_map<ulong, DisplayMode*>::const_iterator it = topmodes.begin(); it != topmodes.end(); ++it)
		sortedmodes.push_back(it->second);
	std::sort(sortedmodes.begin(), sortedmodes.end(), mode_comp());

	// fill the list
	for(std::vector<DisplayMode*>::const_iterator it = sortedmodes.begin(); it != sortedmodes.end(); ++it)
	{
		std::ostringstream ss;
		ss << (*it)->mode.Width << "x" << (*it)->mode.Height << " (" << (*it)->mode.RefreshRate << " Hz)";
		LRESULT index = SendDlgItemMessage(dialog, IDC_MODES, LB_ADDSTRING, 0, reinterpret_cast<LPARAM>(ss.str().c_str()));
		SendDlgItemMessage(dialog, IDC_MODES, LB_SETITEMDATA, index, (*it)->index);
	}

	// ensure the loader is enabled since everything enumerated ok
	Enable(dialog, true);

	// pick the first mode
	SendDlgItemMessage(dialog, IDC_MODES, LB_SETCURSEL, 0, 0);
	mode = static_cast<long>(SendDlgItemMessage(dialog, IDC_MODES, LB_GETITEMDATA, 0, 0));
}

void GraphicsLoader::Enable(const HWND& dialog, const bool& enable) throw()
{
	EnableWindow(GetDlgItem(dialog, IDOK), enable ? TRUE : FALSE);
	EnableWindow(GetDlgItem(dialog, IDC_MODES), enable ? TRUE : FALSE);
	EnableWindow(GetDlgItem(dialog, IDC_WINDOWED), enable ? TRUE : FALSE);
}

GraphicsLoader::GraphicsLoader(const HINSTANCE& instance_, const ComPtr<IDirect3D8>& direct3d_, const D3DFORMAT& depthStencilFormat_) throw()
  : instance(instance_), 
	direct3d(direct3d_), 
	depthStencilFormat(depthStencilFormat_), 
	adapter(0), 
	fullscreen(false), 
	mode(0)
{
}

bool GraphicsLoader::Show(long& adapter_, D3DPRESENT_PARAMETERS& pp) throw(Exception)
{
	// show the dialog
#pragma warning(disable: 4311)
	INT_PTR result = DialogBoxParam(instance, MAKEINTRESOURCE(IDD_LOADER), 0, DialogProc, reinterpret_cast<DWORD>(this));
#pragma warning(default: 4311)
	if(result == -1)
		throw CreateWin32Exception("Error displaying graphics loader: ", GetLastError());

	// exit early if the user didn't click on OK
	if(result != 0)
		return false;

	// get the selected mode
	D3DDISPLAYMODE dm;
	HRESULT hr = direct3d->EnumAdapterModes(adapter, mode, &dm);
	if(FAILED(hr))
		throw CreateDirectXException("GraphicsLoader::Show(): EnumAdapterModes failed: ", hr);
	
	// store the adapter
	adapter_ = adapter;
	
	// set up the presentation parameters
	pp.BackBufferFormat = D3DFMT_A8R8G8B8;
	pp.BackBufferWidth = dm.Width;
	pp.BackBufferHeight = dm.Height;
	pp.Windowed = fullscreen ? FALSE : TRUE;
	if(fullscreen)
		pp.FullScreen_RefreshRateInHz = dm.RefreshRate;
	pp.EnableAutoDepthStencil = (depthStencilFormat == D3DFMT_UNKNOWN) ? FALSE : TRUE;
	if(pp.EnableAutoDepthStencil)
		pp.AutoDepthStencilFormat = depthStencilFormat;

	// return ok
	return true;
}

} // namespace oe
