/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"

namespace oe {

Matrix4x4::Matrix4x4() throw()
  : e11(_R(1)), e12(_R(0)), e13(_R(0)), e14(_R(0)), 
	e21(_R(0)), e22(_R(1)), e23(_R(0)), e24(_R(0)), 
	e31(_R(0)), e32(_R(0)), e33(_R(1)), e34(_R(0)), 
	e41(_R(0)), e42(_R(0)), e43(_R(0)), e44(_R(1))
{
}

Matrix4x4::Matrix4x4(const real_t& e11_, const real_t& e12_, const real_t& e13_, const real_t& e14_, 
					 const real_t& e21_, const real_t& e22_, const real_t& e23_, const real_t& e24_, 
					 const real_t& e31_, const real_t& e32_, const real_t& e33_, const real_t& e34_, 
					 const real_t& e41_, const real_t& e42_, const real_t& e43_, const real_t& e44_) throw()
  : e11(e11_), e12(e12_), e13(e13_), e14(e14_), 
	e21(e21_), e22(e22_), e23(e23_), e24(e24_), 
	e31(e31_), e32(e32_), e33(e33_), e34(e34_), 
	e41(e41_), e42(e42_), e43(e43_), e44(e44_)
{
}

Matrix4x4::Matrix4x4(const real_t& e11_, const real_t& e12_, const real_t& e13_, 
					 const real_t& e21_, const real_t& e22_, const real_t& e23_, 
					 const real_t& e31_, const real_t& e32_, const real_t& e33_) throw()
  : e11(e11_), e12(e12_), e13(e13_), e14(_R(0)), 
	e21(e21_), e22(e22_), e23(e23_), e24(_R(0)), 
	e31(e31_), e32(e32_), e33(e33_), e34(_R(0)), 
	e41(_R(0)), e42(_R(0)), e43(_R(0)), e44(_R(1))
{
}

void Matrix4x4::CreateIdentity() throw()
{
	e11 = e22 = e33 = e44 = _R(1);
	e12 = e13 = e14 = e41 =
	e21 = e23 = e24 = e42 = 
	e31 = e32 = e34 = e43 = _R(0);
}

Matrix4x4::Matrix4x4(const Matrix4x4& m) throw()
  : e11(m.e11), e12(m.e12), e13(m.e13), e14(m.e14), 
	e21(m.e21), e22(m.e22), e23(m.e23), e24(m.e24), 
	e31(m.e31), e32(m.e32), e33(m.e33), e34(m.e34), 
	e41(m.e41), e42(m.e42), e43(m.e43), e44(m.e44)
{
}

Matrix4x4& Matrix4x4::operator=(const Matrix4x4& m) throw()
{
	e11 = m.e11; e12 = m.e12; e13 = m.e13; e14 = m.e14;
	e21 = m.e21; e22 = m.e22; e23 = m.e23; e24 = m.e24;
	e31 = m.e31; e32 = m.e32; e33 = m.e33; e34 = m.e34;
	e41 = m.e41; e42 = m.e42; e43 = m.e43; e44 = m.e44;
	return *this;
}

Matrix4x4 Matrix4x4::Transpose() const throw()
{
	return Matrix4x4(e11, e21, e31, e41, 
					 e12, e22, e32, e42, 
					 e13, e23, e33, e43, 
					 e14, e24, e34, e44);
}

Matrix4x4::Matrix4x4(const Quaternion& q) throw()
{
	CreateFromQuaternion(q);
}

void Matrix4x4::CreateFromQuaternion(const Quaternion& q) throw()
{
	real_t x2 = q.x + q.x;
	real_t y2 = q.y + q.y;
	real_t z2 = q.z + q.z;
	// fill in the non-diagonal entries first
	e12 = e21 = x2 * q.y;
	e23 = e32 = y2 * q.z;
	e31 = e13 = z2 * q.x;
	real_t t = q.w * x2;
	e23 -= t; e32 += t;
	t = q.w * y2;
	e31 -= t; e13 += t;
	t = q.w * z2;
	e12 -= t; e21 += t;
	// fill in the diagonals
	e11 = e22 = e33 = _R(1);
	t = x2 * q.x;
	e22 -= t; e33 -= t;
	t = y2 * q.y;
	e11 -= t; e33 -= t;
	t = z2 * q.z;
	e11 -= t; e22 -= t;
	// fill in the rest
	e14 = e24 = e34 = e41 = e42 = e43 = _R(0);
	e44 = _R(1);
}

void Matrix4x4::CreateFromScale(const Vector3& v) throw()
{
	e11 = v.x; e22 = v.y; e33 = v.z; e44 = _R(1);
	e12 = e13 = e14 = e41 =
	e21 = e23 = e24 = e42 = 
	e31 = e32 = e34 = e43 = _R(0);
}

void Matrix4x4::PreTranslate(const Vector3& v) throw()
{
	e14 += (e11 * v.x + e12 * v.y + e13 * v.z);
	e24 += (e21 * v.x + e22 * v.y + e23 * v.z);
	e34 += (e31 * v.x + e32 * v.y + e33 * v.z);
}

void Matrix4x4::PostTranslate(const Vector3& v) throw()
{
	e14 += v.x;
	e24 += v.y;
	e34 += v.z;
}

void Matrix4x4::PreScale(const Vector3& v) throw()
{
	e11 *= v.x; e12 *= v.y; e13 *= v.z;
	e21 *= v.x; e22 *= v.y; e23 *= v.z;
	e31 *= v.x; e32 *= v.y; e33 *= v.z;
	e41 *= v.x; e42 *= v.y; e43 *= v.z;
}

void Matrix4x4::PostScale(const Vector3& v) throw()
{
	e11 *= v.x; e12 *= v.x; e13 *= v.x; e14 *= v.x;
	e21 *= v.y; e22 *= v.y; e23 *= v.y; e24 *= v.y;
	e31 *= v.z; e32 *= v.z; e33 *= v.z; e34 *= v.z;
}

Vector3 Matrix4x4::GetTranslation() const throw()
{
	return Vector3(e14, e24, e34);
}

Matrix4x4 operator*(const Matrix4x4& l, const Matrix4x4& r) throw()
{
	return Matrix4x4(
		l.e11 * r.e11 + l.e12 * r.e21 + l.e13 * r.e31 + l.e14 * r.e41, 
		l.e11 * r.e12 + l.e12 * r.e22 + l.e13 * r.e32 + l.e14 * r.e42, 
		l.e11 * r.e13 + l.e12 * r.e23 + l.e13 * r.e33 + l.e14 * r.e43, 
		l.e11 * r.e14 + l.e12 * r.e24 + l.e13 * r.e34 + l.e14 * r.e44, 

		l.e21 * r.e11 + l.e22 * r.e21 + l.e23 * r.e31 + l.e24 * r.e41, 
		l.e21 * r.e12 + l.e22 * r.e22 + l.e23 * r.e32 + l.e24 * r.e42, 
		l.e21 * r.e13 + l.e22 * r.e23 + l.e23 * r.e33 + l.e24 * r.e43, 
		l.e21 * r.e14 + l.e22 * r.e24 + l.e23 * r.e34 + l.e24 * r.e44, 

		l.e31 * r.e11 + l.e32 * r.e21 + l.e33 * r.e31 + l.e34 * r.e41, 
		l.e31 * r.e12 + l.e32 * r.e22 + l.e33 * r.e32 + l.e34 * r.e42, 
		l.e31 * r.e13 + l.e32 * r.e23 + l.e33 * r.e33 + l.e34 * r.e43, 
		l.e31 * r.e14 + l.e32 * r.e24 + l.e33 * r.e34 + l.e34 * r.e44, 

		l.e41 * r.e11 + l.e42 * r.e21 + l.e43 * r.e31 + l.e44 * r.e41, 
		l.e41 * r.e12 + l.e42 * r.e22 + l.e43 * r.e32 + l.e44 * r.e42, 
		l.e41 * r.e13 + l.e42 * r.e23 + l.e43 * r.e33 + l.e44 * r.e43, 
		l.e41 * r.e14 + l.e42 * r.e24 + l.e43 * r.e34 + l.e44 * r.e44
	);
}

Vector3 operator*(const Matrix4x4& l, const Vector3& r) throw()
{
	return TransformAs3x3(l, r) + l.GetTranslation();
}

Vector3 TransformAs3x3(const Matrix4x4& l, const Vector3& r) throw()
{
	return Vector3(l.e11 * r.x + l.e12 * r.y + l.e13 * r.z, 
				   l.e21 * r.x + l.e22 * r.y + l.e23 * r.z, 
				   l.e31 * r.x + l.e32 * r.y + l.e33 * r.z);
}

} // namespace oe
