/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#pragma once

namespace oe {

//! A 4x4 matrix class.
class Matrix4x4
{
public:
    //! A matrix component.
	real_t e11, e12, e13, e14;
    //! A matrix component.
	real_t e21, e22, e23, e24;
    //! A matrix component.
    real_t e31, e32, e33, e34;
    //! A matrix component.
    real_t e41, e42, e43, e44;

    //! Creates the identity matrix.
    /*! Creates the identity matrix:
        \f$ \left(
            \begin{array}{cccc}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 1 & 0 \\
                0 & 0 & 0 & 1
            \end{array}
        \right) \f$ .
    */
	Matrix4x4() throw();
    
    //! Creates a matrix with the given components.
	Matrix4x4(const real_t& e11_, const real_t& e12_, const real_t& e13_, const real_t& e14_, 
			  const real_t& e21_, const real_t& e22_, const real_t& e23_, const real_t& e24_, 
			  const real_t& e31_, const real_t& e32_, const real_t& e33_, const real_t& e34_, 
			  const real_t& e41_, const real_t& e42_, const real_t& e43_, const real_t& e44_) throw();

    //! Creates a matrix with the given components.
    /*! Creates a matrix with the given components. The rest of the matrix is filled in with:
        \f$ \left(
            \begin{array}{cccc}
                e_{11} & e_{12} & e_{13} & 0 \\
                e_{21} & e_{22} & e_{23} & 0 \\
                e_{31} & e_{32} & e_{33} & 0 \\
                0 & 0 & 0 & 1
            \end{array}
        \right) \f$ .
    */
    Matrix4x4(const real_t& e11_, const real_t& e12_, const real_t& e13_, 
			  const real_t& e21_, const real_t& e22_, const real_t& e23_, 
			  const real_t& e31_, const real_t& e32_, const real_t& e33_) throw();
	
    //! Creates the identity matrix.
    void CreateIdentity() throw();

    //! Creates a copy of an existing matrix.
	Matrix4x4(const Matrix4x4& m) throw();

    //! Copies an existing matrix.
    Matrix4x4& operator=(const Matrix4x4& m) throw();

    //! Creates a rotation matrix from the given quaternion.
    /*! Creates a rotation matrix from the given quaternion, using the formula: 
        \f$ \left(
            \begin{array}{cccc}
                1 - 2q_{x}^2 & 2(q_x q_y - q_w q_z) & 2(q_x q_z + q_w q_y) & 0 \\
                2(q_x q_y + q_w q_z) & 1 - 2q_{y}^2 & 2(q_y q_z - q_w q_x) & 0 \\
                2(q_x q_z - q_w q_y) & 2(q_y q_z + q_w q_x) & 1 - 2q_{z}^2 & 0 \\
                0 & 0 & 0 & 1
            \end{array}
        \right) \f$ .
    */
	explicit Matrix4x4(const Quaternion& q) throw();

    //! Creates a rotation matrix from the given quaternion.
    /*! Creates a rotation matrix from the given quaternion, using the formula: 
        \f$ \left(
            \begin{array}{cccc}
                1 - 2q_{x}^2 & 2(q_x q_y - q_w q_z) & 2(q_x q_z + q_w q_y) & 0 \\
                2(q_x q_y + q_w q_z) & 1 - 2q_{y}^2 & 2(q_y q_z - q_w q_x) & 0 \\
                2(q_x q_z - q_w q_y) & 2(q_y q_z + q_w q_x) & 1 - 2q_{z}^2 & 0 \\
                0 & 0 & 0 & 1
            \end{array}
        \right) \f$ .
    */
	void CreateFromQuaternion(const Quaternion& q) throw();

    //! Creates a scaling matrix from the given scale.
    /*! Creates a scaling matrix from the given scale, of the form:
        \f$ \left(
            \begin{array}{cccc}
                v_x & 0 & 0 & 0 \\
                0 & v_y & 0 & 0 \\
                0 & 0 & v_z & 0 \\
                0 & 0 & 0 & 1
            \end{array}
        \right) \f$ .
    */
	void CreateFromScale(const Vector3& v) throw();

	//! Applies a translation to the matrix before the existing transformation.
    void PreTranslate(const Vector3& v) throw();
	//! Applies a translation to the matrix after the existing transformation.
	void PostTranslate(const Vector3& v) throw();

	//! Applies a scaling to the matrix before the existing transformation.
	void PreScale(const Vector3& v) throw();
	//! Applies a scaling to the matrix after the existing transformation.
	void PostScale(const Vector3& v) throw(); 

	//! Returns the transpose of this matrix.
	Matrix4x4 Transpose() const throw();

	//! Gets the translational component.
	/*! Gets the translational component, which is the vector \f$ (e_{14}, e_{24}, e_{34}) \f$ .
    */
	Vector3 GetTranslation() const throw();
};

Matrix4x4 operator*(const Matrix4x4& l, const Matrix4x4& r) throw();    //!< Binary matrix multiplication.

//! Transforms a 3-vector as a 4-vector.
Vector3 operator*(const Matrix4x4& l, const Vector3& r) throw();

//! Transforms a 3-vector using the 3x3 sub-matrix.
Vector3 TransformAs3x3(const Matrix4x4& l, const Vector3& r) throw();

} // namespace oe
