/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"

namespace oe {

Quaternion::Quaternion() throw() : w(_R(1)), x(_R(0)), y(_R(0)), z(_R(0)) {}

Quaternion::Quaternion(const real_t& w_, const real_t& x_, const real_t& y_, const real_t& z_) throw()
  : w(w_), x(x_), y(y_), z(z_) 
{
}

Quaternion::Quaternion(const Vector3& axis, const real_t& angle)
{
	CreateFromAxisAngle(axis, angle);
}

Quaternion::Quaternion(const Matrix4x4& m) throw()
{
	CreateFromMatrix(m);
}

void Quaternion::CreateFromMatrix(const Matrix4x4& m) throw()
{
	// get w from the trace
	real_t tmp = m.e11 + m.e22 + m.e33 + _R(1);
	if(tmp < _R(0)) tmp = _R(0);
	w = RealSqrt(tmp) / _R(2);
	// get x, y, z from the body
	tmp	= w * _R(4);
	x = (m.e32 - m.e23) / tmp;
	y = (m.e13 - m.e31) / tmp;
	z = (m.e21 - m.e12) / tmp;
}

real_t Quaternion::GetNorm() const throw()
{
	return RealSqrt(GetNormSquared());
}

real_t Quaternion::GetNormSquared() const throw()
{
	return w * w + x * x + y * y + z * z;
}

Quaternion Quaternion::operator~() const throw()
{
	return Quaternion(w, -x, -y, -z);
}

const Quaternion& Quaternion::operator/=(const real_t& r)
{
	w /= r;
	x /= r;
	y /= r;
	z /= r;
	return *this;
}

void Quaternion::CreateFromAxisAngle(const Vector3& axis, const real_t& angle) throw()
{
	// cache angle / 2
	w = angle * _R(0.5);
	// cache sin(angle / 2)
	x = RealSin(w);
	y = x * axis.y;
	z = x * axis.z;
	x *= axis.x;
	// store cos(angle / 2)
	w = RealCos(w);
}

Quaternion::Quaternion(const Vector3& lookat) throw()
{
	CreateFromLookAtPoint(lookat);
}

void Quaternion::CreateFromLookAtPoint(const Vector3& lookat) throw()
{
	static const real_t lengthEpsilon = _R(0.001); // 1mm reasonable?
	// get the y axis rotation
	real_t r = RealSqrt(lookat.x * lookat.x + lookat.z * lookat.z);
	real_t ycos2 = _R(1), ysin2 = _R(0);
	if(r > lengthEpsilon)
	{
		real_t c = lookat.z / r;
		ycos2 = RealSqrt((_R(1) + c) / 2);
		ysin2 = RealSqrt((_R(1) - c) / 2);
		if(lookat.x < _R(0))
			ysin2 = -ysin2;
	}
	// get the (rotated frame) x axis rotation
	real_t d = RealSqrt(r * r + lookat.y * lookat.y);
	real_t xcos2 = _R(1), xsin2 = _R(0);
	if(d > lengthEpsilon)
	{
		real_t c = r / d;
		xcos2 = RealSqrt((_R(1) + c) / 2);
		xsin2 = RealSqrt((_R(1) - c) / 2);
		if(lookat.y > _R(0))
			xsin2 = -xsin2;
	}
	// TODO: setup the rotation manually
	*this = Quaternion(ycos2, _R(0), ysin2, _R(0)) * Quaternion(xcos2, xsin2, _R(0), _R(0));
}

Quaternion operator*(const Quaternion& l, const Quaternion& r) throw()
{
	return Quaternion(l.w * r.w - (l.x * r.x + l.y * r.y + l.z * r.z), 
					   l.w * r.x + l.x * r.w + l.y * r.z - l.z * r.y,
					   l.w * r.y + l.y * r.w + l.z * r.x - l.x * r.z,
					   l.w * r.z + l.z * r.w + l.x * r.y - l.y * r.x);
}

Quaternion operator/(const Quaternion& l, const real_t& r) throw()
{
	return Quaternion(l.w / r, l.x / r, l.y / r, l.z / r);
}

Quaternion UnitQuaternionPower(const Quaternion& q, const real_t& t) throw()
{
	real_t sinphi = RealSqrt(q.x * q.x + q.y * q.y + q.z * q.z);
	if(sinphi < RealEpsilon)
		return Quaternion();
	// inverse the sin and return the power
	real_t phi = ((q.w < _R(0)) ? -t : t) * RealInvSin(sinphi);
	real_t scale = RealSin(phi) / sinphi;
	return Quaternion(RealCos(phi), q.x * scale, q.y * scale, q.z * scale);
}

Quaternion UnitQuaternionSLERP(const Quaternion& a, const Quaternion& b, const real_t& t) throw()
{
	return UnitQuaternionPower(b * ~a, t) * a;
}

Vector3 UnitQuaternionRotation(const Quaternion& q, const Vector3& v) throw()
{
	// form ([0,v].~q) directly
	Quaternion t(
		v.x * q.x + v.y * q.y + v.z * q.z, 
		v.x * q.w + v.z * q.y - v.y * q.z,
		v.y * q.w + v.x * q.z - v.z * q.x, 			
		v.z * q.w + v.y * q.x - v.x * q.y
	); 
	// return the last three components of (q.[0,v].~q) as a vector
	return Vector3(
	   q.w * t.x + q.x * t.w + q.y * t.z - q.z * t.y,
	   q.w * t.y + q.y * t.w + q.z * t.x - q.x * t.z,
	   q.w * t.z + q.z * t.w + q.x * t.y - q.y * t.x
	);
}

} // namespace oe
