/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#pragma once

namespace oe {

//! A quaternion class.
class Quaternion
{
public:
    //! A quaternion component.
	real_t w, x, y, z;

    //! Creates the unit quaternion.
    /*! Creates the unit quaternion \f$ (1, 0, 0, 0) \f$ .
    */
	Quaternion() throw();
    
    //! Creates a quaternion using the given components.
	Quaternion(const real_t& w_, const real_t& x_, const real_t& y_, const real_t& z_) throw();
    
	//! Create a unit quaternion from the given unit axis and angle.
	/*! Create a unit quaternion from the given unit axis and angle. The quaternion is constructed 
        using the formula:
        \f$ (\cos{\frac{\theta}{2}}, v_x \sin{\frac{\theta}{2}}, 
        v_y \sin{\frac{\theta}{2}}, v_z \sin{\frac{\theta}{2}}) \f$ for \a axis
        \f$ v \f$ and \a angle \f$ \theta \f$ .
    */
    Quaternion(const Vector3& axis, const real_t& angle) throw();

	//! Create a unit quaternion from the given unit axis and angle.
	/*! Create a unit quaternion from the given unit axis and angle. The quaternion is constructed 
        using the formula:
        \f$ (\cos{\frac{\theta}{2}}, v_x \sin{\frac{\theta}{2}}, 
        v_y \sin{\frac{\theta}{2}}, v_z \sin{\frac{\theta}{2}}) \f$ for \a axis
        \f$ v \f$ and \a angle \f$ \theta \f$ .
    */
	void CreateFromAxisAngle(const Vector3& axis, const real_t& angle) throw();

    //! Create a unit quaternion given a unit look direction.
    /*! Create a unit quaternion given a unit look direction. The represented rotation will
        rotate the vector \f$ (0, 0, 1) \f$ to the vector in \a lookat.
    */
	explicit Quaternion(const Vector3& lookat) throw();

    //! Create a unit quaternion given a unit look direction.
    /*! Create a unit quaternion given a unit look direction. The represented rotation will
        rotate the vector \f$ (0, 0, 1) \f$ to the vector in \a lookat.
    */
	void CreateFromLookAtPoint(const Vector3& lookat) throw();

	//! Create a unit quaternion from a rotation matrix.
	/*! Create a unit quaternion from a rotation matrix, using the formula:
    \f$ (\frac{t}{2}, \frac{m_{32} - m_{23}}{2t}, \frac{m_{13} - m_{31}}{2t}, \frac{m_{21} - m_{12}}{2t}) \f$
    where \f$ t = \sqrt{m_{11} + m_{22} + m_{33} + 1} \f$ .
    */
    explicit Quaternion(const Matrix4x4& m) throw();

	//! Create a unit quaternion from a rotation matrix.
	/*! Create a unit quaternion from a rotation matrix, using the formula:
    \f$ (\frac{t}{2}, \frac{m_{32} - m_{23}}{2t}, \frac{m_{13} - m_{31}}{2t}, \frac{m_{21} - m_{12}}{2t}) \f$
    where \f$ t = \sqrt{m_{11} + m_{22} + m_{33} + 1} \f$ .
    */
	void CreateFromMatrix(const Matrix4x4& m) throw();

    //! Compute the Euclidean norm.
    /*! Compute the Euclidean norm: \f$ \sqrt{w^2 + x^2 + y^2 + z^2} \f$ .
    */
	real_t GetNorm() const throw();

    //! Compute the Euclidean norm squared.
    /*! Compute the Euclidean norm squared: \f$ w^2 + x^2 + y^2 + z^2 \f$ .
        \note This operation is considerably cheaper than computing the norm since
        it does not involve the costly square root.
    */
	real_t GetNormSquared() const throw();
	
	//! Unary conjugation.
	/*! Unary conjugation.
        \return the quaternion \f$ (w, -x, -y, -z) \f$ .
    */
    Quaternion operator~() const throw();

	const Quaternion& operator/=(const real_t& r);    //!< Unary scalar division.
};

Quaternion operator*(const Quaternion& l, const Quaternion& r) throw(); //!< Binary quaternion multiplication
Quaternion operator/(const Quaternion& l, const real_t& r) throw();     //!< Binary scalar division

//! Rotates a vector given a unit quaternion.
Vector3 UnitQuaternionRotation(const Quaternion& q, const Vector3& v) throw();

//! Computes a real power of a unit quaternion.
Quaternion UnitQuaternionPower(const Quaternion& q, const real_t& t) throw();

//! Spherical-linear interpolation of unit quaternions.
Quaternion UnitQuaternionSLERP(const Quaternion& a, const Quaternion& b, const real_t& t) throw();

} // namespace oe
