/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"

namespace oe {

AABB::AABB(const Vector3& min_, const Vector3& max_) throw() : min(min_), max(max_) {}

void SpotLight::CreateTransforms() throw()
{
	static const real_t zScale = _R(0x00ffffff - 1);
	static const real_t zNear = _R(1);
	const real_t zFar = range;

	// the spotlight view transform
	Matrix4x4 spotTransform(orientation);
	spotTransform.PreTranslate(-position);

	// the spotlight view projection
	Matrix4x4 projection;
	projection.e11 = fovRatio;
	projection.e22 = fovRatio * aspectRatio;
	projection.e33 = zFar / (zFar - zNear);
	projection.e34 = -projection.e33 * zNear;
	projection.e43 = _R(1);
	projection.e44 = _R(0);

	// the shadow map texture projection
	Matrix4x4 postProjection;
	postProjection.e11 = _R(0.5);
	postProjection.e22 = _R(-0.5);
	postProjection.e14 = postProjection.e24 = _R(0.5);
	postProjection.e33 = zScale;

	shadowMapProjection = postProjection * projection * spotTransform;
    shadowMapTargetProjection = projection * spotTransform;
	
	// the spotlight texture projection
	projection.CreateIdentity();
	projection.e11 = _R(0.5) * fovRatio;
	projection.e22 = _R(-0.5) * fovRatio * aspectRatio;
	projection.e13 = projection.e23 = _R(0.5);

	spotProjection = projection * spotTransform;

	attenuationTransform.CreateIdentity();
	attenuationTransform.PostTranslate(-position);
	real_t invrange = _R(0.5) / range;
	attenuationTransform.PostScale(Vector3(invrange, invrange, invrange));
	attenuationTransform.PostTranslate(Vector3(_R(0.5), _R(0.5), _R(0.5)));
}

const Matrix4x4& SpotLight::GetAttenuationTransform() const throw()
{
	return attenuationTransform;
}

const Matrix4x4& SpotLight::GetShadowMapTargetProjection() const throw()
{
	return shadowMapTargetProjection;
}

const Matrix4x4& SpotLight::GetShadowMapProjection() const throw()
{
	return shadowMapProjection;
}

const Matrix4x4& SpotLight::GetSpotProjection() const throw()
{
	return spotProjection;
}

Matrix4x4 SpotLight::GetWorldLightTransform() const throw()
{
	return Matrix4x4(
		_R(-1), _R(0), _R(0), position.x, 
		_R(0), _R(-1), _R(0), position.y, 
		_R(0), _R(0), _R(-1), position.z, 
		_R(0), _R(0), _R(0), _R(1)
	);
}

Matrix4x4 SpotLight::GetViewLightTransform(const Matrix4x4& viewTransform) const throw()
{
	Vector3 viewPosition = viewTransform * position;
	return Matrix4x4(
		_R(-1), _R(0), _R(0), viewPosition.x, 
		_R(0), _R(-1), _R(0), viewPosition.y, 
		_R(0), _R(0), _R(-1), viewPosition.z, 
		_R(0), _R(0), _R(0), _R(1)
	) * viewTransform;
}

Renderable::Renderable(const AABB& aabb_, 
					   const Material& material_, 
					   const OrientedVertexBuffer& vertices_, 
					   const TexCoordBuffer& diffuseTexCoords_) throw()
  : aabb(aabb_), 
	material(material_), 
	vertices(vertices_), 
	diffuseTexCoords(diffuseTexCoords_)
{
}

Renderable::~Renderable() throw() {}

const AABB& Renderable::GetAABB() const throw()
{
	return aabb;
}

const Material& Renderable::GetMaterial() const throw()
{
	return material;
}

const OrientedVertexBuffer& Renderable::GetVertices() const throw()
{
	return vertices;
}

const TexCoordBuffer& Renderable::GetDiffuseTexCoords() const throw()
{
	return diffuseTexCoords;
}

const Matrix4x4& Renderable::GetTransform() const throw()
{
	return transform;
}

const Quaternion& Renderable::GetOrientation() const throw()
{
	return orientation;
}

} // namespace oe