/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#pragma once

namespace oe {

//! Oriented vertex with normal and tangent component.
struct OrientedVertex
{
	Vector3 p;	//!< Position.
	Vector3 n;	//!< Unit normal.
	Vector3 t;	//!< Unit tangent.
};

//! A vertex buffer of oriented vertices.
typedef VertexBuffer<OrientedVertex> OrientedVertexBuffer;

//! 2D texture coordinate.
struct TexCoord
{
	//! Texture coordinate.
	float u, v;
};

//! A vertex buffer of 2D texture coordinates.
typedef VertexBuffer<TexCoord> TexCoordBuffer;

//! Axis-aligned bounding box.
class AABB
{
public:
	//! A box extremum.
	Vector3 min, max;

	//! Creates a new bounding box.
	AABB(const Vector3& min_, const Vector3& max_) throw();
};

//! A simple spotlight class.
class SpotLight
{
private:
	Matrix4x4 attenuationTransform;
	Matrix4x4 shadowMapTargetProjection;
	Matrix4x4 shadowMapProjection;
	Matrix4x4 spotProjection;

public:
	Vector3 colour;				//!< The light colour.
	Texture intensity;			//!< The intensity as a projected texture.
	VolumeTexture attenuation;	//!< The attenuation as a volume texture.
	
	Vector3 position;			//!< The position.
	Quaternion orientation;		//!< The orientation.
	real_t fovRatio;			//!< The field-of-view ratio.
	real_t aspectRatio;			//!< The aspect ratio.
	real_t range;				//!< The maximum range of the light.

	Texture rendertarget;		//!< A surface to use as the render target when creating the shadow map.
	Texture shadowmap;			//!< The shadow map depth surface.

	//! Updates all the light transforms based on the new light position and orientation.
	void CreateTransforms() throw();
	//! Gets the transform from world space to attenuation space.
	const Matrix4x4& GetAttenuationTransform() const throw();
	//! Gets the transform from world space to the target shadow map projection.
	const Matrix4x4& GetShadowMapTargetProjection() const throw();
	//! Gets the transform from world space to the shadow map texture space.
	const Matrix4x4& GetShadowMapProjection() const throw();
	//! Gets the transform from world space to the spotlight projection.
	const Matrix4x4& GetSpotProjection() const throw();
	//! Computes the world space to light direction transform.
	Matrix4x4 GetWorldLightTransform() const throw();
	//! Computes the world space to light direction in view space transform.
	Matrix4x4 GetViewLightTransform(const Matrix4x4& viewTransform) const throw();
};

//! A base class for renderable entities.
class Renderable
{
protected:
	AABB aabb;							//!< A bounding box for the entity.
	Material material;					//!< The material to use for the mesh.
	OrientedVertexBuffer vertices;		//!< The vertex buffer containing the mesh vertices.
	TexCoordBuffer diffuseTexCoords;	//!< The vertex buffer containing the mesh diffuse texture coordinates.
	Matrix4x4 transform;				//!< The local transform for the mesh vertices from object to world space.
	Quaternion orientation;				//!< The local orientation of the mesh vertices from object to world space.

public:
	//! Creates a renderable entity with the given parameters.
	Renderable(const AABB& aabb_, const Material& material_, const OrientedVertexBuffer& vertices_, const TexCoordBuffer& diffuseTexCoords_) throw();

	//! Virtual since it will likely be deleted through a pointer to this class.
	virtual ~Renderable() throw();

	//! Gets the bounding box for this entity.
	const AABB& GetAABB() const throw();
	//! Gets the material for this entity.
	const Material& GetMaterial() const throw();
	//! Gets the vertex buffer that contains this entity's vertices.
	const OrientedVertexBuffer& GetVertices() const throw();
	//! Gets the vertex buffer that contains this entity's diffuse texture coordinates.
	const TexCoordBuffer& GetDiffuseTexCoords() const throw();
	//! Gets the local transform for this entity.
	const Matrix4x4& GetTransform() const throw();
	//! Gets the local orientation of this entity.
	const Quaternion& GetOrientation() const throw();

	//! Renders this entity.
	virtual void DrawPrimitive(const Graphics& graphics) const throw(Exception) = 0;
};

} // namespace oe