/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#pragma once

namespace oe {

//! Unlocks a locked section of a vertex or index buffer when no longer referenced.
template<typename T, typename I>
class SafeLockDeleter
{
private:
	ComPtr<I> buffer;

public:
	//! Creates a new deleter for a particular device.
	explicit SafeLockDeleter(const ComPtr<I>& buffer_) throw();

	//! Deletes the handle using the encapsulated device.
	void operator()(T* const&) throw();
};

//! Exception-safe wrapper around a locked section of a vertex or index buffer.
template<typename T, typename I>
class SafeLock : public SharedHandle<T*, SafeLockDeleter<T, I> >
{
public:
	//! Creates a new safe lock around the given memory location and buffer.
	SafeLock(T* const& lock, const ComPtr<I>& buffer) throw();

	//! Explicitly unlocks the buffer.
	/*! Explicitly unlocks the buffer. Calling this method is usually unnecessary since the 
		buffer is unlocked automatically when the last reference to the lock goes out of scope.
	*/
	void Unlock() throw();

	//! Space-saving index into the locked buffer.
	/*! Space-saving index into the locked buffer.
		\todo	Locks know their size and check for access violations?
	*/
	T& operator[](const ulong& index) throw(Exception);
};

} // namespace oe

// Template implementation
#include "SafeLock.inl"

