/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"

namespace oe {

#ifndef DOXYGEN_PARSING
namespace scene {
#include "ambientVS.inl"
#include "ambientPS.inl"
#include "textureVS.inl"
#include "texturePS.inl"
#include "attenprojVS.inl"
#include "attenprojPS.inl"
#include "attenprojshadowVS.inl"
#include "attenprojshadowPS.inl"
#include "pointbumpdiffuseVS.inl"
#include "pointbumpdiffusePS.inl"
#include "pointbumpreflectVS.inl"
#include "pointbumpreflectPS.inl"
}
#endif // ndef DOXYGEN_PARSING

//! The vertex shader declaration for vertices with position only.
static ulong VertexDecl[] = {
	D3DVSD_STREAM(0),
	D3DVSD_REG(0, D3DVSDT_FLOAT3), // position
	D3DVSD_SKIP(6), // skip the normal and tangent
	D3DVSD_END()
};

//! The vertex shader declaration for a stream pair of position only and texture coordinates.
static ulong TexVertexDecl[] = {
	D3DVSD_STREAM(0),
	D3DVSD_REG(0, D3DVSDT_FLOAT3), // position
	D3DVSD_SKIP(6), // skip the normal and tangent
	D3DVSD_STREAM(1),
	D3DVSD_REG(3, D3DVSDT_FLOAT2), // texture co-ordinate
	D3DVSD_END()
};

//! The vertex shader declaration for a stream pair of oriented vertices and texture coordinates.
static ulong OrientedTexVertexDecl[] = {
	D3DVSD_STREAM(0),
	D3DVSD_REG(0, D3DVSDT_FLOAT3), // position
	D3DVSD_REG(1, D3DVSDT_FLOAT3), // normal
	D3DVSD_REG(2, D3DVSDT_FLOAT3), // tangent
	D3DVSD_STREAM(1),
	D3DVSD_REG(3, D3DVSDT_FLOAT2), // texture co-ordinate
	D3DVSD_END()
};

SceneEntity::SceneEntity(const boost::shared_ptr<Renderable>& entity_) throw() : entity(entity_) {}

SceneFragment::SceneFragment(const long& pass_, 
							 const OE_SCENESHADER& shader_, 
							 const boost::weak_ptr<Renderable>& entity_) throw()
  :	pass(pass_), 
	shader(shader_), 
	entity(entity_)
{
}

SceneFragment::SceneFragment(const long& pass_, 
							 const OE_SCENESHADER& shader_,
							 const boost::weak_ptr<Renderable>& entity_, 
							 const boost::weak_ptr<SpotLight>& light_) throw()
  :	pass(pass_), 
	shader(shader_), 
	entity(entity_), 
	light(light_)
{
}

inline bool operator<(const SceneFragment& l, const SceneFragment& r) throw()
{
	// order by pass first
	if(l.pass != r.pass)
		return l.pass < r.pass;

	// then by shader
	if(l.shader != r.shader)
		return l.shader < r.shader;

	// then by other stuff
	return false;
}

Scene::Scene(const Graphics& graphics) throw(Exception) : entityPasses(0), stateChanges(0)
{
	// create all the shaders
	ambientVS = graphics.CreateVertexShader(VertexDecl, scene::dwAmbientVertexShader);
	ambientPS = graphics.CreatePixelShader(scene::dwAmbientPixelShader);
	textureVS = graphics.CreateVertexShader(TexVertexDecl, scene::dwTextureVertexShader);
	texturePS = graphics.CreatePixelShader(scene::dwTexturePixelShader);
	attenuatedProjVS = graphics.CreateVertexShader(VertexDecl, scene::dwAttenprojVertexShader);
	attenuatedProjPS = graphics.CreatePixelShader(scene::dwAttenprojPixelShader);
	attenuatedProjShadowedVS = graphics.CreateVertexShader(VertexDecl, scene::dwAttenprojshadowVertexShader);
	attenuatedProjShadowedPS = graphics.CreatePixelShader(scene::dwAttenprojshadowPixelShader);
	pointBumpDiffuseVS = graphics.CreateVertexShader(OrientedTexVertexDecl, scene::dwPointbumpdiffuseVertexShader);
	pointBumpDiffusePS = graphics.CreatePixelShader(scene::dwPointbumpdiffusePixelShader);
	pointBumpReflectVS = graphics.CreateVertexShader(OrientedTexVertexDecl, scene::dwPointbumpreflectVertexShader);
	pointBumpReflectPS = graphics.CreatePixelShader(scene::dwPointbumpreflectPixelShader);

	// create the attenuation volume texture
	static const ulong volumeLength = 64;
	HRESULT hr = graphics->CreateVolumeTexture(volumeLength, volumeLength, volumeLength, 1, 0, 
		D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, attenuation.AddressOf());
	if(FAILED(hr))
		throw CreateDirectXException("Scene::Scene(): CreateVolumeTexture failed: ", hr);
	ComputeAttenuationVolumeIndirect(graphics, attenuation);

	// create the normalisation cube texture
	static const ulong normalisationLength = 128;
	hr = graphics->CreateCubeTexture(normalisationLength, 1, 0, D3DFMT_X8R8G8B8, D3DPOOL_DEFAULT, normalisation.AddressOf());
	if(FAILED(hr))
		throw CreateDirectXException("Scene::Scene(): CreateCubeTexture failed: ", hr);
	ComputeCubeTextureIndirect(graphics, normalisation, NormalisationEvaluator());
}

void Scene::AddRenderable(const boost::shared_ptr<Renderable>& entity) throw(Exception)
{
	// check it has a diffuse map
    if(!entity->GetMaterial().diffuse)
		throw Exception("Scene::AddRenderable(): Entity has no diffuse texture.");
	// add the entity to the scene
	scene.push_back(SceneEntity(entity));
}

void Scene::AddSpotLight(const boost::shared_ptr<SpotLight>& spotlight) throw(Exception)
{
	spotlights.push_back(spotlight);
}

void Scene::RenderShadowMaps(const Graphics& graphics) throw(Exception)
{
	// store a reference to the current render surfaces
	Surface rendertarget, depthstencil;
	HRESULT hr = graphics->GetRenderTarget(rendertarget.AddressOf());
	if(FAILED(hr))
		throw CreateDirectXException("Scene::RenderShadowMaps(): GetRenderTarget failed: ", hr);
	hr = graphics->GetDepthStencilSurface(depthstencil.AddressOf());
	if(FAILED(hr))
		throw CreateDirectXException("Scene::RenderShadowMaps(): GetRenderTarget failed: ", hr);

	// render the shadow maps
	graphics.SetVertexShader(ambientVS);
	graphics.SetPixelShader(ambientPS);
	graphics->SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);
	graphics->SetRenderState(D3DRS_COLORWRITEENABLE, 0);
	for(SpotLights::const_iterator it = spotlights.begin(); it != spotlights.end(); ++it)
	{
		boost::weak_ptr<SpotLight> light(*it);
	
		// get the shadow map dimensions
		D3DSURFACE_DESC desc;
		hr = light.get()->rendertarget->GetLevelDesc(0, &desc);
		if(FAILED(hr))
			throw CreateDirectXException("Scene::Render(): GetLevelDesc failed: ", hr);
		
		// set up for rendering to this shadow map
		Surface spottarget, spotdepth;
		hr = light.get()->rendertarget->GetSurfaceLevel(0, spottarget.AddressOf());
		if(FAILED(hr))
			throw CreateDirectXException("Scene::Render(): GetSurfaceLevel failed: ", hr);
		hr = light.get()->shadowmap->GetSurfaceLevel(0, spotdepth.AddressOf());
		if(FAILED(hr))
			throw CreateDirectXException("Scene::Render(): GetSurfaceLevel failed: ", hr);
        
		hr = graphics->SetRenderTarget(spottarget.Get(), spotdepth.Get());
		if(FAILED(hr))
			throw CreateDirectXException("Scene::RenderShadowMaps(): SetRenderTarget failed: ", hr);

		graphics->Clear(0, 0, D3DCLEAR_ZBUFFER, 0, 1.0f, 0);
		graphics->BeginScene();

		Matrix4x4 texProjection;
		texProjection.PostTranslate(Vector3(_R(-1) / _R(desc.Width), _R(1) / _R(desc.Height), _R(0)));

		Matrix4x4 shadowProjection;
		static const real_t zNear = _R(1);
		const real_t zFar = light.get()->range;
		shadowProjection.e11 = light.get()->fovRatio;
		shadowProjection.e22 = light.get()->fovRatio * light.get()->aspectRatio;
		shadowProjection.e33 = zFar / (zFar - zNear);
		shadowProjection.e34 = -shadowProjection.e33 * zNear;
		shadowProjection.e43 = _R(1);
		shadowProjection.e44 = _R(0);
		
		Matrix4x4 lightTransform(light.get()->orientation);
		lightTransform.PreTranslate(-light.get()->position);

		const float blue[] = { 1.0f, 1.0f, 1.0f, 1.0f };
		graphics->SetPixelShaderConstant(PS_LIGHTCOLOUR, blue, 1);

		Matrix4x4 objectProjection;
		for(Entities::const_iterator it = scene.begin(); it != scene.end(); ++it)
		{
			boost::weak_ptr<Renderable> entity(it->entity);

			objectProjection = texProjection * shadowProjection * lightTransform * entity.get()->GetTransform();
			graphics->SetVertexShaderConstant(VS_PROJECTION, &objectProjection, 4);

			graphics.SetStreamSource(0, entity.get()->GetVertices());
			entity.get()->DrawPrimitive(graphics);
			graphics.ReleaseStreamSource(0);
		}

		graphics->EndScene();
	}
	graphics->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
	graphics->SetRenderState(D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_RGBA);

	// restore the surfaces
	hr = graphics->SetRenderTarget(rendertarget.Get(), depthstencil.Get());
	if(FAILED(hr))
		throw CreateDirectXException("Scene::RenderShadowMaps(): SetRenderTarget failed: ", hr);
}

void Scene::SetView(const Vector3& position, const Quaternion& orientation, const Projection& projection) throw(Exception)
{
	// set the up the view transformations
	viewOrientation = orientation;
	viewTransform.CreateFromQuaternion(orientation);
	viewTransform.PreTranslate(-position);
	viewProjection = projection;

	// make the list of fragments
	fragments.clear();
	for(Entities::const_iterator it = scene.begin(); it != scene.end(); ++it)
	{
		// ambient pass
        fragments.push_back(SceneFragment(0, OESS_AMBIENTLIGHT, it->entity));

		// diffuse light
		fragments.push_back(SceneFragment(1, OESS_ATTENUATEDPROJECTEDSHADOWEDINTENSITY, it->entity, *spotlights.begin()));
		fragments.push_back(SceneFragment(2, OESS_BUMPEDDIFFUSEPOINTLIGHT, it->entity, *spotlights.begin()));
		fragments.push_back(SceneFragment(3, OESS_DIFFUSERESPONSE, it->entity));

		// specular light
		fragments.push_back(SceneFragment(4, OESS_BUMPEDSPECULARPOINTLIGHT, it->entity, *spotlights.begin()));
	}
}

void Scene::Compile() throw()
{
	// sort the scene fragments
	std::sort(fragments.begin(), fragments.end(), std::less<SceneFragment>());
}

void Scene::BeginShader(const Graphics& graphics, const OE_SCENESHADER& shader) const throw()
{
	switch(shader)
	{
	case OESS_AMBIENTLIGHT:
		graphics.SetVertexShader(ambientVS);
		graphics.SetPixelShader(ambientPS);
		graphics->SetRenderState(D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_RGBA);
		graphics->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
		graphics->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);
		graphics->SetRenderState(D3DRS_ZFUNC, D3DCMP_LESSEQUAL);
		break;
	case OESS_ATTENUATEDPROJECTEDSHADOWEDINTENSITY:
		graphics.SetVertexShader(attenuatedProjShadowedVS);
		graphics.SetPixelShader(attenuatedProjShadowedPS);
		graphics->SetRenderState(D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_ALPHA);
		graphics->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
		graphics->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
		graphics->SetRenderState(D3DRS_ZFUNC, D3DCMP_EQUAL);
		graphics->SetTextureStageState(0, D3DTSS_ADDRESSU, D3DTADDRESS_CLAMP);
		graphics->SetTextureStageState(0, D3DTSS_ADDRESSV, D3DTADDRESS_CLAMP);
		graphics->SetTextureStageState(1, D3DTSS_ADDRESSU, D3DTADDRESS_CLAMP);
		graphics->SetTextureStageState(1, D3DTSS_ADDRESSV, D3DTADDRESS_CLAMP);
		graphics->SetTextureStageState(1, D3DTSS_ADDRESSW, D3DTADDRESS_CLAMP);
		graphics.SetTexture(1, attenuation);
		break;
	case OESS_BUMPEDDIFFUSEPOINTLIGHT:
		graphics.SetVertexShader(pointBumpDiffuseVS);
		graphics.SetPixelShader(pointBumpDiffusePS);
		graphics->SetRenderState(D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_RGB);
		graphics->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
		graphics->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_DESTALPHA);
		graphics->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
		graphics->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
		graphics->SetRenderState(D3DRS_ZFUNC, D3DCMP_EQUAL);
		graphics.SetTexture(1, normalisation);
		break;
	case OESS_DIFFUSERESPONSE:
		graphics.SetVertexShader(textureVS);
		graphics.SetPixelShader(texturePS);
		graphics->SetRenderState(D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_RGB);
		graphics->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
		graphics->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_DESTCOLOR);
		graphics->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ZERO);
		graphics->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
		graphics->SetRenderState(D3DRS_ZFUNC, D3DCMP_EQUAL);
		break;
	case OESS_BUMPEDSPECULARPOINTLIGHT:
		graphics.SetVertexShader(pointBumpReflectVS);
		graphics.SetPixelShader(pointBumpReflectPS);
		graphics->SetRenderState(D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_RGB);
		graphics->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
		graphics->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_DESTALPHA);
		graphics->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
		graphics->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
		graphics->SetRenderState(D3DRS_ZFUNC, D3DCMP_EQUAL);
		break;
	}
}

void Scene::EndShader(const Graphics& graphics, const OE_SCENESHADER& shader) const throw()
{
	switch(shader)
	{
	case OESS_AMBIENTLIGHT:
		break;
	case OESS_ATTENUATEDPROJECTEDSHADOWEDINTENSITY:
		graphics.ReleaseTexture(1);
		graphics->SetTextureStageState(1, D3DTSS_ADDRESSU, D3DTADDRESS_WRAP);
		graphics->SetTextureStageState(1, D3DTSS_ADDRESSV, D3DTADDRESS_WRAP);
		graphics->SetTextureStageState(1, D3DTSS_ADDRESSW, D3DTADDRESS_WRAP);
		graphics->SetTextureStageState(0, D3DTSS_ADDRESSU, D3DTADDRESS_WRAP);
		graphics->SetTextureStageState(0, D3DTSS_ADDRESSV, D3DTADDRESS_WRAP);
		break;
	case OESS_BUMPEDDIFFUSEPOINTLIGHT:
		graphics.ReleaseTexture(1);
		break;
	case OESS_DIFFUSERESPONSE:
		break;
	case OESS_BUMPEDSPECULARPOINTLIGHT:
		break;
	case OESS_NOTDEFINED:
		graphics->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
		graphics->SetRenderState(D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_RGBA);
		graphics->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);
		graphics->SetRenderState(D3DRS_ZFUNC, D3DCMP_LESSEQUAL);
		break;
	}
}

void Scene::Render(const Graphics& graphics) throw(Exception)
{
	entityPasses = stateChanges = 0;

	// render the scene
	graphics->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
	graphics->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
	graphics->SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_LINEAR);
	graphics->SetTextureStageState(1, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
	graphics->SetTextureStageState(1, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
	graphics->SetTextureStageState(1, D3DTSS_MIPFILTER, D3DTEXF_LINEAR);
	graphics->SetTextureStageState(2, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
	graphics->SetTextureStageState(2, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
	graphics->SetTextureStageState(2, D3DTSS_MIPFILTER, D3DTEXF_LINEAR);
	graphics->SetTextureStageState(3, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
	graphics->SetTextureStageState(3, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
	graphics->SetTextureStageState(3, D3DTSS_MIPFILTER, D3DTEXF_LINEAR);

	const Matrix4x4 worldProjection = viewProjection.GetTransform() * viewTransform;
	const Vector3 ambientLight(_R(0.1), _R(0.1), _R(0.1));

	OE_SCENESHADER current = OESS_NOTDEFINED;
	for(Fragments::const_iterator it = fragments.begin(); it != fragments.end(); ++it, ++entityPasses)
	{
		// switch shader if necessary
		if(it->shader != current)
		{
			EndShader(graphics, current);
			current = it->shader;
			BeginShader(graphics, current);
			++stateChanges;
		}
		Renderable* entity = it->entity.get();
		SpotLight* light = it->light.get();
		// load the constants and resources for this entity
		switch(current)
		{
		case OESS_AMBIENTLIGHT:
			graphics.SetPixelShaderConstant(PS_LIGHTCOLOUR, ambientLight);
			break;
		case OESS_ATTENUATEDPROJECTEDSHADOWEDINTENSITY:
			graphics.SetVertexShaderConstant(VS_SPOTTRANSFORM, light->GetSpotProjection() * entity->GetTransform(), 3);
			graphics.SetVertexShaderConstant(VS_ATTENUATIONTRANSFORM, light->GetAttenuationTransform() * entity->GetTransform(), 3);
			graphics.SetVertexShaderConstant(VS_SHADOWTRANSFORM, light->GetShadowMapProjection() * entity->GetTransform());
			graphics.SetTexture(0, light->intensity);
			graphics.SetTexture(2, light->shadowmap);
			break;
		case OESS_BUMPEDDIFFUSEPOINTLIGHT:
            graphics.SetPixelShaderConstant(PS_LIGHTCOLOUR, entity->GetMaterial().diffuseIntensity * light->colour);
			graphics.SetVertexShaderConstant(VS_LIGHTWORLDTRANSFORM, light->GetWorldLightTransform() * entity->GetTransform());
			graphics.SetVertexShaderConstant(VS_WORLDORIENTATION, Matrix4x4(entity->GetOrientation()), 3);
			graphics.SetStreamSource(1, entity->GetDiffuseTexCoords());
			graphics.SetTexture(0, entity->GetMaterial().normals);
			break;
		case OESS_DIFFUSERESPONSE:
			graphics.SetStreamSource(1, entity->GetDiffuseTexCoords());
			graphics.SetTexture(0, entity->GetMaterial().diffuse);
			break;
		case OESS_BUMPEDSPECULARPOINTLIGHT:
			graphics.SetPixelShaderConstant(PS_LIGHTCOLOUR, entity->GetMaterial().specularIntensity * light->colour);
			graphics.SetVertexShaderConstant(VS_LIGHTVIEWTRANSFORM, light->GetViewLightTransform(viewTransform) * entity->GetTransform());
			graphics.SetVertexShaderConstant(VS_VIEWORIENTATION, Matrix4x4(viewOrientation * entity->GetOrientation()), 3);
			graphics.SetStreamSource(1, entity->GetDiffuseTexCoords());
			graphics.SetTexture(0, entity->GetMaterial().normals);
			graphics.SetTexture(3, entity->GetMaterial().specular);
			break;
		}
		// render
		graphics.SetVertexShaderConstant(VS_PROJECTION, worldProjection * entity->GetTransform());
		graphics.SetStreamSource(0, entity->GetVertices());
		entity->DrawPrimitive(graphics);
		graphics.ReleaseStreamSource(0);
		// unload the resources for this entity
		switch(current)
		{
		case OESS_AMBIENTLIGHT:
			break;
		case OESS_ATTENUATEDPROJECTEDSHADOWEDINTENSITY:
			graphics.ReleaseTexture(2);
			graphics.ReleaseTexture(0);
			break;
		case OESS_BUMPEDDIFFUSEPOINTLIGHT:
			graphics.ReleaseTexture(0);
			graphics.ReleaseStreamSource(1);
			break;
		case OESS_DIFFUSERESPONSE:
			graphics.ReleaseTexture(0);
			graphics.ReleaseStreamSource(1);
			break;
		case OESS_BUMPEDSPECULARPOINTLIGHT:
			graphics.ReleaseTexture(3);
			graphics.ReleaseTexture(0);
			graphics.ReleaseStreamSource(1);
			break;
		}
	}
	EndShader(graphics, current);
	current = OESS_NOTDEFINED;
	BeginShader(graphics, current);
}

const ulong& Scene::GetNumEntityPasses() const throw()
{
	return entityPasses;
}

const ulong& Scene::GetNumStateChanges() const throw()
{
	return stateChanges;
}

} // namespace oe