/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#pragma once

namespace oe {

//! An enumeration of all the shaders the scene manager knows about.
enum OE_SCENESHADER
{
	OESS_NOTDEFINED,
	OESS_AMBIENTLIGHT,
	OESS_ATTENUATEDPROJECTEDSHADOWEDINTENSITY,
	OESS_BUMPEDDIFFUSEPOINTLIGHT,
	OESS_DIFFUSERESPONSE,
	OESS_BUMPEDSPECULARPOINTLIGHT
};
	
//! Used internally to represent a scene fragment to be rendered using a particular shader.
class SceneFragment
{
public:
	long pass;							//!< The scene pass of this fragment.
	OE_SCENESHADER shader;				//!< The shader to use.
	boost::weak_ptr<Renderable> entity; //!< The entity to render.
	boost::weak_ptr<SpotLight> light;	//!< The light to use to light it (can be null).

public: 
	//! Creates a new scene fragment.
	SceneFragment(const long& pass_, const OE_SCENESHADER& shader_, const boost::weak_ptr<Renderable>& entity_) throw();

	//! Creates a new scene fragment.
	SceneFragment(const long& pass_, const OE_SCENESHADER& shader_, const boost::weak_ptr<Renderable>& entity_, const boost::weak_ptr<SpotLight>& light_) throw();
};

//! Implements an ordering on the fragments so that they can be sorted.
bool operator<(const SceneFragment& l, const SceneFragment& r) throw();

//! Used to hold a reference to a scene entity.
/*! Used to hold a reference to a scene entity. This will have more stuff in it 
	once I start making the renderer more robust.
*/
class SceneEntity
{
public:
	boost::shared_ptr<Renderable> entity;	//!< The renderable object of this entity.

public:
	//! Creates a new scene entity.
	SceneEntity(const boost::shared_ptr<Renderable>& entity_) throw();
};

//! A scene graph object.
class Scene
{
private:
	typedef std::list<SceneEntity> Entities;
	typedef std::list<boost::shared_ptr<SpotLight> > SpotLights;
	typedef std::vector<SceneFragment> Fragments;

	Entities scene;
	SpotLights spotlights;
	
	VertexShader ambientVS, textureVS, attenuatedProjVS, attenuatedProjShadowedVS, pointBumpDiffuseVS, pointBumpReflectVS;
	PixelShader ambientPS, texturePS, attenuatedProjPS, attenuatedProjShadowedPS, pointBumpDiffusePS, pointBumpReflectPS;
	VolumeTexture attenuation;
	CubeTexture normalisation;

	Fragments fragments;
	Matrix4x4 viewTransform;
	Quaternion viewOrientation;
	Projection viewProjection;

	ulong entityPasses;
	ulong stateChanges;

	void BeginShader(const Graphics& graphics, const OE_SCENESHADER& shader) const throw();
	void EndShader(const Graphics& graphics, const OE_SCENESHADER& shader) const throw();
	
public:
	//! Creates an empty scene graph.
	Scene(const Graphics& graphics) throw(Exception);

	//! Add a new renderable object to the scene.
	void AddRenderable(const boost::shared_ptr<Renderable>& entity) throw(Exception);
	//! Add a new spotlight to the scene.
	void AddSpotLight(const boost::shared_ptr<SpotLight>& spotlight) throw(Exception);

	//! Render the shadow maps for all spotlights.	
	void RenderShadowMaps(const Graphics& graphics) throw(Exception);
	
	//! Set the viewpoint to render from.
	void SetView(const Vector3& position, const Quaternion& orientation, const Projection& projection) throw(Exception);
	//! Compile the scene to perform the sorting and batching.
	void Compile() throw();
	//! Renders the scene
	void Render(const Graphics& graphics) throw(Exception);

	//! Gets the number of entity passes (i.e. fragments) used to render the scene.
	const ulong& GetNumEntityPasses() const throw();
	//! Gets the number of state changes needed to render the scene.
	const ulong& GetNumStateChanges() const throw();
};

} // namespace oe