/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#pragma once

namespace oe {

//! A template shared resource handler.
template<typename H, typename Deleter>
class SharedHandle
{
private:
	H handle;
	ulong* count;
	Deleter deleter;
	
	void Reset(const SharedHandle<H, Deleter>& sh) throw();

protected:
	//! Frees this shared handle.
	/*! Frees this shared handle. If this is the last share then the deleter is used
		to delete the resource.
	*/
	void Free() throw();

public:
	//! Creates an empty resource handler.
	SharedHandle() throw();
	
	//! Destroys the encapsulated resource if this is the last reference to it.
	~SharedHandle() throw();
	
	//! Creates a new resource handler around the given resource.
	SharedHandle(const H& handle_, const Deleter& deleter_) throw();
	
	//! Creates a copy an existing resource handle.
	SharedHandle(const SharedHandle<H, Deleter>& sh) throw();
	
	//! Copies an existing resource handle.
	SharedHandle<H, Deleter>& operator=(const SharedHandle<H, Deleter>& sh) throw();
	
	//! Returns a reference to the encapsulated handle. 
	const H& Get() const throw();
	
	template<typename U, typename V>
	friend std::ostream& operator<<(std::ostream& os, const SharedHandle<U, V>& sh) throw();	
};

//! Outputs the handle value and reference count for debugging.
template<typename U, typename V>
std::ostream& operator<<(std::ostream& os, const SharedHandle<U, V>& sh) throw();	

} // namespace oe

// All template implementation
#include "SharedHandle.inl"
