/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"

namespace oe {

TextBuffer::TextBuffer(const Graphics& graphics, const ulong& capacity_) throw(Exception)
  : capacity(capacity_), 
	used(0), 
	x(0), 
	y(0)
{
	vertices = graphics.CreateVertexBuffer<GuiVertex>(6 * capacity, OEBT_DYNAMIC);
}

void TextBuffer::Clear() throw()
{
	used = 0;
}

void TextBuffer::SetPosition(const short& x_, const short& y_) throw()
{
	x = x_; 
	y = y_;
}

void TextBuffer::SetFont(const Font& font_) throw()
{
	font = font_;
	Clear();
}

void TextBuffer::Println(const std::string& text) throw(Exception)
{
	// check we have  font
	const uchar height = font.GetHeight();
	if(height == 0)
		throw Exception("TextBuffer::Println(): No font set.");

	// get the text and length
	ulong length = static_cast<ulong>(text.length());
	if(used + length > capacity)
		length = capacity - used;

	// lock the buffer and fill
	VertexBuffer<GuiVertex>::SafeLock lock = vertices.Lock(used * 6, length * 6);

	const char* in = text.c_str();
	GuiVertex* out = lock.Get();
	short pos = x;
	const real_t scale = _R(1) / _R(16 * height);

	for(ulong i = 0; i < length; ++i)
	{
		const uchar c = *in++;
		const uchar width = font.GetWidth(c);

		// set the vertices
		out[0].x = out[1].x = out[3].x = pos;
		out[2].x = out[4].x = out[5].x = pos + width;
		out[1].y = out[2].y = out[4].y = y;
		out[0].y = out[3].y = out[5].y = y + font.GetHeight();

		// set the texture coordinates, remembering to 
		// offset by 0.5 texels for pixel-perfect alignment
		long u = (c & 0x0F) * height;
		long v = ((c >> 4) & 0x0F) * height;
		out[0].u = out[1].u = out[3].u = (_R(u) + _R(0.5)) * scale;
		out[2].u = out[4].u = out[5].u = (_R(u + width) + _R(0.5)) * scale;
		out[1].v = out[2].v = out[4].v = (_R(v) + _R(0.5)) * scale; 
		out[0].v = out[3].v = out[5].v = (_R(v + height) + _R(0.5)) * scale; 

		// next!
		pos = pos + width;
		out += 6;
	}

	used += length;
	y = y + height;
}

void TextBuffer::Println() throw()
{
	y = y + font.GetHeight();
}

void TextBuffer::DrawPrimitive(const Graphics& graphics) const throw(Exception)
{
	if(used)
	{
		graphics.SetTexture(0, font.GetTexture());
		graphics.SetStreamSource(0, vertices);
		graphics->DrawPrimitive(D3DPT_TRIANGLELIST, 0, used * 2);
		graphics.ReleaseStreamSource(0);
		graphics.ReleaseTexture(0);
	}
}

} // namespace oe
