/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"

namespace oe {

Vector3::Vector3() throw() : x(_R(0)), y(_R(0)), z(_R(0)) {}

Vector3::Vector3(const real_t& x_, const real_t& y_, const real_t& z_) throw() : x(x_), y(y_), z(z_) {}

Vector3::Vector3(const Vector3& v) throw() : x(v.x), y(v.y), z(v.z) {}

Vector3& Vector3::operator=(const Vector3& v) throw()
{
	x = v.x;
	y = v.y;
	z = v.z;
	return *this;
}

real_t Vector3::GetNorm() const throw()
{
	return RealSqrt(GetNormSquared());
}

real_t Vector3::GetNormSquared() const throw()
{
	return x * x + y * y + z * z;
}

Vector3 Vector3::operator-() const throw()
{
	return Vector3(-x, -y, -z);
}

const Vector3& Vector3::operator+=(const Vector3& v) throw()
{
	x += v.x;
	y += v.y;
	z += v.z;
	return *this;
}

const Vector3& Vector3::operator-=(const Vector3& v) throw()
{
	x -= v.x;
	y -= v.y;
	z -= v.z;
	return *this;
}

const Vector3& Vector3::operator*=(const Vector3& v) throw()
{
	x *= v.x;
	y *= v.y; 
	z *= v.z;
	return *this;
}

const Vector3& Vector3::operator*=(const real_t& r) throw()
{
	x *= r;
	y *= r;
	z *= r;
	return *this;
}

const Vector3& Vector3::operator/=(const real_t& r) throw()
{
	x /= r;
	y /= r;
	z /= r;
	return *this;
}

Vector3 operator+(const Vector3& l, const Vector3& r) throw()
{
	return Vector3(l.x + r.x, l.y + r.y, l.z + r.z);
}

Vector3 operator-(const Vector3& l, const Vector3& r) throw()
{
	return Vector3(l.x - r.x, l.y - r.y, l.z - r.z);
}

Vector3 operator*(const Vector3& l, const Vector3& r) throw()
{
	return Vector3(l.x * r.x, l.y * r.y, l.z * r.z);
}

Vector3 operator*(const real_t& l, const Vector3& r) throw()
{
	return Vector3(l * r.x, l * r.y, l * r.z);
}

Vector3 operator*(const Vector3& l, const real_t& r) throw()
{
	return Vector3(l.x * r, l.y * r, l.z * r);
}

Vector3 operator/(const Vector3& l, const real_t& r) throw()
{
	return Vector3(l.x / r, l.y / r, l.z / r);
}

real_t DotProduct(const Vector3& l, const Vector3& r) throw()
{
	return l.x * r.x + l.y * r.y + l.z * r.z;
}

Vector3 CrossProduct(const Vector3& l, const Vector3& r) throw()
{
	return Vector3(l.y * r.z - l.z * r.y, 
				   l.z * r.x - l.x * r.z, 
				   l.x * r.y - l.y * r.x);
}

std::ostream& operator<<(std::ostream& os, const Vector3& v) throw()
{
	return os << v.x << ", " << v.y << ", " << v.z;
}

} // namespace oe
