/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#pragma once

namespace oe {

//! A 3-vector class.
/*! A 3-vector class. Provides overloaded operators for common mathematical functions.
*/
class Vector3
{
public:
    //! A vector component.
	real_t x, y, z;

public:
    //! Creates the zero 3-vector.
	Vector3() throw();

    //! Creates a 3-vector with the given components.
	Vector3(const real_t& x_, const real_t& y_, const real_t& z_) throw();

    //! Creates a copy of an existing 3-vector.
	Vector3(const Vector3& v) throw();
    
    //! Copies an existing 3-vector.
    /*! Copies an existing 3-vector.
        \param  v The vector to copy.
        \return A reference to this 3-vector.
    */
	Vector3& operator=(const Vector3& v) throw();

    //! Compute the Euclidean norm.
    /*! Compute the Euclidean norm: \f$ \sqrt{x^2 + y^2 + z^2} \f$ .
    */
	real_t GetNorm() const throw();
    
    //! Compute the Euclidean norm squared.
    /*! Compute the Euclidean norm squared: \f$ x^2 + y^2 + z^2 \f$ .
        \note This operation is considerably cheaper than computing the norm since
        it does not involve the costly square root.
    */
	real_t GetNormSquared() const throw();

	Vector3 operator-() const throw();                    //!< Unary negation.

	const Vector3& operator+=(const Vector3& v) throw();  //!< Unary addition
	const Vector3& operator-=(const Vector3& v) throw();  //!< Unary subtraction.
	const Vector3& operator*=(const Vector3& v) throw();  //!< Unary multiplication

	const Vector3& operator*=(const real_t& r) throw();   //!< Unary scalar multiplication.
	const Vector3& operator/=(const real_t& r) throw();   //!< Unary scalar division.
};

Vector3 operator+(const Vector3& l, const Vector3& r) throw();	//!< Binary addition.
Vector3 operator-(const Vector3& l, const Vector3& r) throw();	//!< Binary component-wise subtraction.
Vector3 operator*(const Vector3& l, const Vector3& r) throw();  //!< Binary component-wise multiplication.
Vector3 operator*(const real_t& l, const Vector3& r) throw();   //!< Binary scalar multiplication.
Vector3 operator*(const Vector3& l, const real_t& r) throw();   //!< Binary scalar multiplication.
Vector3 operator/(const Vector3& l, const real_t& r) throw();   //!< Binary scalar division.

//! Computes the dot product.
/*! Computes the dot product using the standard formula
    \f$ a \cdot b = a_x b_x + a_y b_y + a_z b_z \f$ .
*/
real_t DotProduct(const Vector3& l, const Vector3& r) throw();

//! Computes the cross product.
/*! Computes the cross product using the standard formula
    \f$ a \wedge b = (a_y b_z - a_z b_y,  a_z b_x - a_x b_z, a_x b_y - a_y b_x) \f$ .
*/
Vector3 CrossProduct(const Vector3& l, const Vector3& r) throw();

//! Outputs a vector to a std::ostream.
std::ostream& operator<<(std::ostream& os, const Vector3& v) throw();

} // namespace oe
