/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"

namespace oe {

inline void ReportException(const Exception& e) throw()
{
	MessageBox(0, e.GetExcuse().c_str(), "OpenEngine: Committing harakiri...", MB_OK | MB_TASKMODAL | MB_ICONERROR);
}

//! Creates the device and calls oe_main guarded by catch blocks.
static void InnerMain(const HINSTANCE& instance) throw()
{
	// Employ a 2-level catch so that any run-time exceptions are displayed before 
	// the message loop gets nuked by WM_QUIT when the Graphics class destroys itself.
	try
	{
		// create the directx wrappers
		Graphics graphics(instance);
		try
		{
			// run the application
			oe_main(graphics);

			// log exit
			std::cout << "Application ended." << std::endl;
		}
		catch(const Exception& e) { ReportException(e); }
	}
	catch(const Exception& e) 
	{ 
		ReportException(e); 
	}
	catch(const std::exception& e) 
	{ 
		std::string msg("std::exception: ");
		ReportException(Exception(msg.append(e.what()))); 
	}
	catch(...) 
	{ 
		ReportException(Exception("Unknown exception caught!")); 
	}
}

} // namespace oe

#ifdef DOXYGEN_PARSING
// present so that this function gets documented
void oe_main(oe::Graphics& graphics) throw(oe::Exception) { graphics; }
#endif

int APIENTRY WinMain(HINSTANCE instance, HINSTANCE /*prevInstance*/, LPSTR /*cmdLine*/, int /*cmdShow*/)
{
	// open a log file for writing
	std::ofstream log("out.txt");
	if(log.fail())
	{
		MessageBox(0, "Failed to open \"log.txt\" for writing.", 
			"OpenEngine: Committing harakiri...", MB_OK | MB_TASKMODAL | MB_ICONERROR
		);
		return 0xdeadbeef;
	}

	// redirect the standard streams
    std::streambuf *outbuf = std::cout.rdbuf(log.rdbuf());
    std::streambuf *errbuf = std::cerr.rdbuf(log.rdbuf());

	// run
	oe::InnerMain(instance);
	
	// restore the buffers
    std::cout.rdbuf(outbuf);
    std::cerr.rdbuf(errbuf);
	return 0;
}

