/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"

namespace oe {

void CreateCubeTriangleListVertices(void* const& buffer, const ulong& stride) throw()
{	
	size_t offset = reinterpret_cast<size_t>(buffer);
	float* buf;
#define FOO(X, Y, Z)	buf = reinterpret_cast<float*>(offset);			\
						*buf++ = _R(X); *buf++ = _R(Y); *buf = _R(Z);	\
						offset += stride;
    FOO(-1, -1, -1) FOO(-1,  1, -1) FOO( 1,  1, -1)
    FOO( 1,  1, -1) FOO( 1, -1, -1) FOO(-1, -1, -1)
    FOO(-1, -1, -1) FOO(-1, -1,  1) FOO(-1,  1,  1)
    FOO(-1,  1,  1) FOO(-1,  1, -1) FOO(-1, -1, -1)
    FOO(-1, -1, -1) FOO( 1, -1, -1) FOO( 1, -1,  1)
    FOO( 1, -1,  1) FOO(-1, -1,  1) FOO(-1, -1, -1)

	FOO( 1,  1,  1) FOO(-1,  1,  1) FOO(-1, -1,  1)
    FOO(-1, -1,  1) FOO( 1, -1,  1) FOO( 1,  1,  1)
    FOO( 1,  1,  1) FOO( 1, -1,  1) FOO( 1, -1, -1)
    FOO( 1, -1, -1) FOO( 1,  1, -1) FOO( 1,  1,  1)
    FOO( 1,  1,  1) FOO( 1,  1, -1) FOO(-1,  1, -1)
    FOO(-1,  1, -1) FOO(-1,  1,  1) FOO( 1,  1,  1)
#undef FOO
}

void CreateCubeTriangleListNormals(void* const& buffer, const ulong& stride) throw()
{
	size_t offset = reinterpret_cast<size_t>(buffer);
	float* buf;
#define BAR(X, Y, Z)	for(ulong n = 0; n < 6; ++n)						\
						{													\
							buf = reinterpret_cast<float*>(offset);			\
							*buf++ = _R(X); *buf++ = _R(Y); *buf = _R(Z);	\
							offset += stride;								\
						}
	BAR(0, 0, -1)
	BAR(-1, 0, 0)
	BAR(0, -1, 0)

	BAR(0, 0, 1)
	BAR(1, 0, 0)
	BAR(0, 1, 0)
#undef BAR
}

void CreateCubeTriangleListTangents(void* const& buffer, const ulong& stride) throw()
{
	size_t offset = reinterpret_cast<size_t>(buffer);
	float* buf;
#define BAR(X, Y, Z)	for(ulong n = 0; n < 6; ++n)						\
						{													\
							buf = reinterpret_cast<float*>(offset);			\
							*buf++ = _R(X); *buf++ = _R(Y); *buf = _R(Z);	\
							offset += stride;								\
						}
	BAR(1, 0, 0)
	BAR(0, 1, 0)
	BAR(0, 0, 1)

	BAR(0, -1, 0)
	BAR(0, 0, -1)
	BAR(-1, 0, 0)
#undef BAR
}

void CreateCubeTriangleListTexcoords(void* const& buffer, const ulong& stride) throw()
{
	size_t offset = reinterpret_cast<size_t>(buffer);
	float* buf;
#define FOO(U, V)	buf = reinterpret_cast<float*>(offset);	\
					*buf++ = _R(U); *buf = _R(V);			\
					offset += stride;
#define BAR(U0, U1, V0, V1)	FOO(U0, V1)	\
							FOO(U0, V0)	\
							FOO(U1, V0)	\
							FOO(U1, V0)	\
							FOO(U1, V1)	\
							FOO(U0, V1)
	for(oe::ulong i = 0; i < 6; ++i)
	{
		BAR(0, 1, 0, 1)
	}
#undef BAR
#undef FOO
}

//! Evaluates a torus at the given parametric coordinate.
static inline Vector3 EvaluateTorus(const real_t& outerRadius, const real_t& innerRadius, const real_t& u, const real_t& v) throw()
{
	return Vector3(
		RealCos(u * _R(2) * RealPi) * (outerRadius - innerRadius * RealSin(v * _R(2) * RealPi)), 
		-RealSin(u * _R(2) * RealPi) * (outerRadius - innerRadius * RealSin(v * _R(2) * RealPi)), 
		-innerRadius * RealCos(v * _R(2) * RealPi) 
	);
}

//! Evaluates a torus normal at the given parametric coordinate.
static inline Vector3 EvaluateTorusNormal(const real_t& u, const real_t& v) throw()
{
	return Vector3(
		-RealCos(u * _R(2) * RealPi) * RealSin(v * _R(2) * RealPi), 
		RealSin(u * _R(2) * RealPi) * RealSin(v * _R(2) * RealPi), 
		-RealCos(v * _R(2) * RealPi)
	);
}

//! Evaluates a torus tangent at the given parametric coordinate.
static inline Vector3 EvaluateTorusTangent(const real_t& u, const real_t& /*v*/) throw()
{
	return Vector3(
		-RealSin(u * _R(2) * RealPi), 
		-RealCos(u * _R(2) * RealPi), 
		_R(0)
	);
}

//! Evaluates a torus binormal at the given parametric coordinate.
static inline Vector3 EvaluateTorusBiNormal(const real_t& u, const real_t& v) throw()
{
	return Vector3(
		-RealCos(u * _R(2) * RealPi) * RealCos(v * _R(2) * RealPi), 
		RealSin(u * _R(2) * RealPi) * RealCos(v * _R(2) * RealPi), 
		RealSin(v * _R(2) * RealPi)
	);
}

//! Evaluates a torus knot at the given parametric coordinate.
static inline Vector3 EvaluateTorusKnot(const real_t& outerRadius, const real_t& innerRadius, 
										const real_t& knotRadius, 
										const real_t& p, const real_t& q, const real_t& lenpq, 
										const real_t& u, const real_t& v) throw()
{
    return EvaluateTorus(outerRadius, innerRadius, p * u, q * u) + knotRadius * 
		(RealCos(v * _R(2) * RealPi) * EvaluateTorusNormal(p * u, q * u)
		+ (q / lenpq) * RealSin(v * _R(2) * RealPi) * EvaluateTorusBiNormal(p * u, q * u)
		- (p / lenpq) * RealSin(v * _R(2) * RealPi) * EvaluateTorusTangent(p * u, q * u));
}

//! Evaluates a torus knot normal at the given parametric coordinate.
static inline Vector3 EvaluateTorusKnotNormal(const real_t& p, const real_t& q, const real_t& lenpq, const real_t& u, const real_t& v) throw()
{
    return RealCos(v * _R(2) * RealPi) * EvaluateTorusNormal(p * u, q * u)
		+ (q / lenpq) * RealSin(v * _R(2) * RealPi) * EvaluateTorusBiNormal(p * u, q * u)
		- (p / lenpq) * RealSin(v * _R(2) * RealPi) * EvaluateTorusTangent(p * u, q * u);
}

//! Evaluates a torus knot tangent at the given parametric coordinate.
static inline Vector3 EvaluateTorusKnotTangent(const real_t& p, const real_t& q, const real_t& lenpq, const real_t& u, const real_t& /*v*/) throw()
{
	return (p / lenpq) * EvaluateTorusTangent(p * u, q * u)
		+ (q / lenpq) * EvaluateTorusBiNormal(p * u, q * u);
}

void CreateTorusVertices(void* const& buffer, const ulong& stride, 
						 const real_t& outerRadius, const real_t& innerRadius, 
						 const ulong outerSegments, const ulong& innerSegments) throw()
{
	size_t offset = reinterpret_cast<size_t>(buffer);
	for(ulong v = 0; v <= innerSegments; ++v)
	{
		for(ulong u = 0; u <= outerSegments; ++u)
		{
			*reinterpret_cast<Vector3*>(offset) = EvaluateTorus(outerRadius, innerRadius, _R(u) / _R(outerSegments), _R(v) / _R(innerSegments));
			offset += stride;
		}
	}
}

void CreateTorusNormals(void* const& buffer, const ulong& stride, 
						const ulong outerSegments, const ulong& innerSegments) throw()
{
	size_t offset = reinterpret_cast<size_t>(buffer);
	for(ulong v = 0; v <= innerSegments; ++v)
	{
		for(ulong u = 0; u <= outerSegments; ++u)
		{
			*reinterpret_cast<Vector3*>(offset) = EvaluateTorusNormal(_R(u) / _R(outerSegments), _R(v) / _R(innerSegments));
			offset += stride;
		}
	}
}

void CreateTorusTangents(void* const& buffer, const ulong& stride, 
						 const ulong outerSegments, const ulong& innerSegments) throw()
{
	size_t offset = reinterpret_cast<size_t>(buffer);
	for(ulong v = 0; v <= innerSegments; ++v)
	{
		for(ulong u = 0; u <= outerSegments; ++u)
		{
			*reinterpret_cast<Vector3*>(offset) = EvaluateTorusTangent(_R(u) / _R(outerSegments), _R(v) / _R(innerSegments));
			offset += stride;
		}
	}
}

void CreateTorusTexcoords(void* const& buffer, const ulong& stride, 
						  const ulong outerSegments, const ulong& innerSegments) throw()
{
	size_t offset = reinterpret_cast<size_t>(buffer);
	for(ulong v = 0; v <= innerSegments; ++v)
	{
		for(ulong u = 0; u <= outerSegments; ++u)
		{
			reinterpret_cast<float*>(offset)[0] = _R(u) / _R(outerSegments);
			reinterpret_cast<float*>(offset)[1] = _R(v) / _R(innerSegments);
			offset += stride;
		}
	}
}

void CreateTorusIndices(ushort* const& buffer, const ushort& offset, 
						const ulong outerSegments, const ulong& innerSegments) throw()
{
	// we'll add one degenerate triangle each side of the strip so that
	// the whole torus can be drawn in one DrawIndexedPrimitive call
	ushort* buf = buffer;
#define FOO(N, M)	*buf++ = offset + static_cast<ushort>(N);					\
					for(ulong i = 0; i <= outerSegments; ++i)					\
					{															\
						*buf++ = offset + static_cast<ushort>(N + i);			\
						*buf++ = offset + static_cast<ushort>(M + i);			\
					}															\
					*buf++ = offset + static_cast<ushort>(M + outerSegments);

	for(ulong v = 0; v < innerSegments;)
	{
        ulong v0 = v++;
		FOO(v0 * (outerSegments + 1), v * (outerSegments + 1))
	}

#undef FOO
}

void CreateTorusKnotVertices(void* const& buffer, const ulong& stride, 
							 const real_t& outerRadius, const real_t& innerRadius, const real_t& knotRadius, 
							 const real_t& p, const real_t& q, 
							 const ulong outerSegments, const ulong& innerSegments) throw()
{
	real_t lenpq = RealSqrt(p * p + q * q);
	size_t offset = reinterpret_cast<size_t>(buffer);
	for(ulong v = 0; v <= innerSegments; ++v)
	{
		for(ulong u = 0; u <= outerSegments; ++u)
		{
			*reinterpret_cast<Vector3*>(offset) = EvaluateTorusKnot(
				outerRadius, innerRadius, knotRadius, 
				p, q, lenpq, _R(u) / _R(outerSegments), _R(v) / _R(innerSegments)
			);
			offset += stride;
		}
	}
}

void CreateTorusKnotNormals(void* const& buffer, const ulong& stride, 
							const real_t& p, const real_t& q, 
							const ulong outerSegments, const ulong& innerSegments) throw()
{
	real_t lenpq = RealSqrt(p * p + q * q);
	size_t offset = reinterpret_cast<size_t>(buffer);
	for(ulong v = 0; v <= innerSegments; ++v)
	{
		for(ulong u = 0; u <= outerSegments; ++u)
		{
			*reinterpret_cast<Vector3*>(offset) = EvaluateTorusKnotNormal(
				p, q, lenpq, 
				_R(u) / _R(outerSegments), _R(v) / _R(innerSegments)
			);
			offset += stride;
		}
	}
}

void CreateTorusKnotTangents(void* const& buffer, const ulong& stride,
							 const real_t& p, const real_t& q,
							 const ulong outerSegments, const ulong& innerSegments) throw()
{
	real_t lenpq = RealSqrt(p * p + q * q);
	size_t offset = reinterpret_cast<size_t>(buffer);
	for(ulong v = 0; v <= innerSegments; ++v)
	{
		for(ulong u = 0; u <= outerSegments; ++u)
		{
			*reinterpret_cast<Vector3*>(offset) = EvaluateTorusKnotTangent(
				p, q, lenpq, 
				_R(u) / _R(outerSegments), _R(v) / _R(innerSegments)
			);
			offset += stride;
		}
	}
}

//! Converts a normal map colour into a unit vector.
static Vector3 NormalMapToVector3(const ulong& value) throw()
{
	Vector3 temp(
		_R((value >> 16) & 0xff) - _R(127.5), 
		_R((value >> 8) & 0xff) - _R(127.5), 
		_R(value & 0xff) - _R(127.5)
	);
	return temp / temp.GetNorm();
}

//! Converts a unit vector into a normal map colour.
static ulong Vector3ToNormalMap(const Vector3& value) throw()
{
	real_t n = _R(1) / value.GetNorm();
	ulong r = static_cast<ulong>((_R(0.5) * value.x * n + _R(0.5)) * _R(255));
	ulong g = static_cast<ulong>((_R(0.5) * value.y * n + _R(0.5)) * _R(255));
	ulong b = static_cast<ulong>((_R(0.5) * value.z * n + _R(0.5)) * _R(255));
	return (r << 16) | (g << 8) | b;
}

//! Filters one normal map surface into a smaller one using a box filter.
static void BoxFilterNormalMap(const Surface& source, const Surface& dest) throw(Exception)
{
	// get the surface dimensions
	D3DSURFACE_DESC sourceDesc, destDesc;
	HRESULT hr = source->GetDesc(&sourceDesc);
	if(FAILED(hr))
		throw CreateDirectXException("BoxFilterNormalMap(): GetDesc failed: ", hr);
	hr = dest->GetDesc(&destDesc);
	if(FAILED(hr))
		throw CreateDirectXException("BoxFilterNormalMap(): GetDesc failed: ", hr);

	// compute the filter box size and check the surface formats
	ulong boxWidth = sourceDesc.Width / destDesc.Width;
	ulong boxHeight = sourceDesc.Height / destDesc.Height;

	if(destDesc.Width * boxWidth != sourceDesc.Width ||
		destDesc.Height * boxHeight != sourceDesc.Height)
		throw Exception("NormalMapBoxFilter(): Incompatible surface dimensions.");
	if(destDesc.Format != sourceDesc.Format || destDesc.Format != D3DFMT_X8R8G8B8)
		throw Exception("NormalMapBoxFilter(): Incompatible surface format.");

	// lock the surfaces
	D3DLOCKED_RECT sourceLock, destLock;
	hr = source->LockRect(&sourceLock, 0, 0);
	if(FAILED(hr))
		throw CreateDirectXException("BoxFilterNormalMap(): LockRect failed: ", hr);
	hr = dest->LockRect(&destLock, 0, 0);
	if(FAILED(hr))
	{
		source->UnlockRect();
		throw CreateDirectXException("BoxFilterNormalMap(): LockRect failed: ", hr);
	}

	// filter
	boost::scoped_array<Vector3> row(new Vector3[destDesc.Width]);
	memset(row.get(), 0, destDesc.Width * sizeof(Vector3));

	ulong destY = 0;
	ulong boxY = 0;
	for(ulong sourceY = 0; sourceY < sourceDesc.Height; ++sourceY)
	{
		// add in this row to the current filtered row
		ulong destX = 0;
		ulong boxX = 0;
		ulong* sourceBuffer = reinterpret_cast<ulong*>(reinterpret_cast<size_t>(sourceLock.pBits) + sourceY * sourceLock.Pitch);
		for(ulong sourceX = 0; sourceX < sourceDesc.Width; ++sourceX)
		{
			row[destX] += NormalMapToVector3(sourceBuffer[sourceX]);
			if(++boxX == boxWidth)
			{
				boxX = 0;
				++destX;
			}
		}

		// if we've filtered a whole child row, write it to the child texture
		if(++boxY == boxHeight)
		{
			ulong* destBuffer = reinterpret_cast<ulong*>(reinterpret_cast<size_t>(destLock.pBits) + destY * destLock.Pitch);
			for(destX = 0; destX < destDesc.Width; ++destX)
				destBuffer[destX] = Vector3ToNormalMap(row[destX]);
			// zero the row
			memset(row.get(), 0, destDesc.Width * sizeof(Vector3));
			boxY = 0;
			++destY;
		}
	}

	// unlock the surfaces
	dest->UnlockRect();
	source->UnlockRect();
}

void FilterNormalMap(const Texture& texture) throw(Exception)
{
	// get the number of texture levels
	const ulong numLevels = texture->GetLevelCount();
	if(numLevels < 2)
		throw Exception("FilterNormalMap(): Texture has no mipmap levels.");

	// get the levels
	boost::scoped_array<Surface> levels(new Surface[numLevels]);
	for(ulong i = 0; i < numLevels; ++i)
	{
		HRESULT hr = texture->GetSurfaceLevel(i, levels[i].AddressOf());
		if(FAILED(hr))
			throw CreateDirectXException("FilterNormapMap(): GetSurfaceLevel failed: ", hr);
	}

	// recursively filter each level
	for(ulong i = 1; i < numLevels; ++i)
		BoxFilterNormalMap(levels[i-1], levels[i]);

	// update the 'dirty' region
	HRESULT hr = texture->AddDirtyRect(0);
	if(FAILED(hr))
		throw CreateDirectXException("FilterNormalMap(): AddDirtyRect failed: ", hr);
}

void ComputeAttenuationVolume(const VolumeTexture& volume) throw(Exception)
{
	// get the dimensions
	D3DVOLUME_DESC desc;
	HRESULT hr = volume->GetLevelDesc(0, &desc);
	if(FAILED(hr))
		throw CreateDirectXException("ComputeAttenuationVolume(): GetLevelDesc failed: ", hr);
	if(desc.Width != desc.Height || desc.Height != desc.Depth)
		throw Exception("ComputeAttenuationVolume(): Non-cubic volumes are not supported.");
	if(desc.Format != D3DFMT_A8R8G8B8)
		throw Exception("ComputeAttenuationVolume(): Unsupported format.");
	// lock the whole volume
	D3DLOCKED_BOX lock;
	hr = volume->LockBox(0, &lock, 0, 0);
	if(FAILED(hr))
		throw CreateDirectXException("ComputeAttenuationVolume(): LockBox failed: ", hr);
	// fill
	const real_t offset = _R(desc.Width - 1) * _R(0.5);
	const real_t offsetsq = offset * offset;
	for(ulong z = 0; z < desc.Width; ++z)
	{
		for(ulong y = 0; y < desc.Width; ++y)
		{
			ulong* buffer = reinterpret_cast<ulong*>(reinterpret_cast<size_t>(lock.pBits) 
				+ z * lock.SlicePitch + y * lock.RowPitch);
			for(ulong x = 0; x < desc.Width; ++x)
			{
				const real_t dx = (_R(x) - offset);
				const real_t dy = (_R(y) - offset);
				const real_t dz = (_R(z) - offset);
				real_t r = _R(1) - _R(dx * dx + dy * dy + dz * dz) / offsetsq;
				if(r > _R(1)) r = _R(1);
				if(r < _R(0)) r = _R(0);
                ulong i = static_cast<ulong>(r * _R(255));
				buffer[x] = (i << 24);
			}
		}
	}
	volume->UnlockBox(0);
}

void ComputeAttenuationVolumeIndirect(const Graphics& graphics, const VolumeTexture& volume) throw(Exception)
{
	// get the dimensions
	D3DVOLUME_DESC desc;
	HRESULT hr = volume->GetLevelDesc(0, &desc);
	if(FAILED(hr))
		throw CreateDirectXException("ComputeAttenuationVolumeIndirect(): GetLevelDesc failed: ", hr);

	// create a system memory version
	VolumeTexture temp;
	hr = graphics->CreateVolumeTexture(desc.Width, desc.Height, desc.Depth, 1, 0, D3DFMT_A8R8G8B8, D3DPOOL_SYSTEMMEM, temp.AddressOf());
	if(FAILED(hr))
		throw CreateDirectXException("ComputeAttenuationVolumeIndirect(): CreateVolumeTexture failed: ", hr);
	
	// compute the volume into the system memory version 
	ComputeAttenuationVolume(temp);

	// copy to the hardware one
	hr = volume->AddDirtyBox(0);
	if(FAILED(hr))
		throw CreateDirectXException("ComputeAttenuationVolumeIndirect(): AddDirtyBox failed: ", hr);
	hr = graphics->UpdateTexture(temp.Get(), volume.Get());
	if(FAILED(hr))
		throw CreateDirectXException("ComputeAttenuationVolumeIndirect(): UpdateTexture failed: ", hr);
}

ulong NormalisationEvaluator::operator()(const Vector3& ray) const throw()
{
	real_t n = _R(1) / ray.GetNorm();
	ulong r = static_cast<ulong>((_R(0.5) * ray.x * n + _R(0.5)) * _R(255));
	ulong g = static_cast<ulong>((_R(0.5) * ray.y * n + _R(0.5)) * _R(255));
	ulong b = static_cast<ulong>((_R(0.5) * ray.z * n + _R(0.5)) * _R(255));
	return (r << 16) | (g << 8) | b;
}

SpecularEvaluator::SpecularEvaluator(const real_t& power_) throw() : power(power_) {}

ulong SpecularEvaluator::operator()(const Vector3& ray) const throw()
{
	static const Vector3 light(_R(0), _R(0), _R(-1));
	real_t r = DotProduct(ray, light) / ray.GetNorm();
	if(r < _R(0)) r = _R(0);
	else if(r > _R(1)) r = _R(1);
	else r = RealPow(r, power);
	ulong c = static_cast<ulong>(r * _R(255));
	return (c << 24);
}

Texture CopyTextureToDefaultPool(const Graphics& graphics, const Texture& texture) throw(Exception)
{
	// get the dimensions
	D3DSURFACE_DESC desc;
	HRESULT hr = texture->GetLevelDesc(0, &desc);
	if(FAILED(hr))
		throw CreateDirectXException("CopyToDefaultPool(): GetLevelDesc failed: ", hr);

	// create a copy
	Texture ret;
	hr = graphics->CreateTexture(desc.Width, desc.Height, 0, 0, desc.Format, D3DPOOL_DEFAULT, ret.AddressOf());
	if(FAILED(hr))
		throw CreateDirectXException("CopyToDefaultPool(): CreateTexture failed: ", hr);

	// update the copy
	hr = ret->AddDirtyRect(0);
	if(FAILED(hr))
		throw CreateDirectXException("CopyToDefaultPool(): AddDirtyRect failed: ", hr);
	hr = graphics->UpdateTexture(texture.Get(), ret.Get());
	if(FAILED(hr))
		throw CreateDirectXException("CopyToDefaultPool(): UpdateTexture failed: ", hr);

	return ret;
}

} // namespace oe