/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#pragma once

namespace oe {

//! Utility function to create the vertices of a cube as a triangle list.
void CreateCubeTriangleListVertices(void* const& buffer, const ulong& stride) throw();
//! Utility function to create the normals of a cube as a triangle list.
void CreateCubeTriangleListNormals(void* const& buffer, const ulong& stride) throw();
//! Utility function to create the tangents of a cube as a triangle list.
void CreateCubeTriangleListTangents(void* const& buffer, const ulong& stride) throw();
//! Utility function to create the texture coordinates of a cube as a triangle list.
void CreateCubeTriangleListTexcoords(void* const& buffer, const ulong& stride) throw();

//! Utility function to create the vertices of a torus as an indexed triangle strip.
void CreateTorusVertices(void* const& buffer, const ulong& stride, const real_t& outerRadius, const real_t& innerRadius, const ulong outerSegments, const ulong& innerSegments) throw();
//! Utility function to create the normals of a torus as an indexed triangle strip.
void CreateTorusNormals(void* const& buffer, const ulong& stride, const ulong outerSegments, const ulong& innerSegments) throw();
//! Utility function to create the tangents of a torus as an indexed triangle strip.
void CreateTorusTangents(void* const& buffer, const ulong& stride, const ulong outerSegments, const ulong& innerSegments) throw();
//! Utility function to create the texture coordinates of a torus as an indexed triangle strip.
void CreateTorusTexcoords(void* const& buffer, const ulong& stride, const ulong outerSegments, const ulong& innerSegments) throw();
//! Utility function to create the indices of a torus as an indexed triangle strip.
void CreateTorusIndices(ushort* const& buffer, const ushort& offset, const ulong outerSegments, const ulong& innerSegments) throw();

//! Utility function to create the vertices of a torus knot as an indexed triangle strip.
void CreateTorusKnotVertices(void* const& buffer, const ulong& stride, const real_t& outerRadius, const real_t& innerRadius, const real_t& knotRadius, const real_t& p, const real_t& q, const ulong outerSegments, const ulong& innerSegments) throw();
//! Utility function to create the normals of a torus knot as an indexed triangle strip.
void CreateTorusKnotNormals(void* const& buffer, const ulong& stride, const real_t& p, const real_t& q, const ulong outerSegments, const ulong& innerSegments) throw();
//! Utility function to create the tangents of a torus knot as an indexed triangle strip.
void CreateTorusKnotTangents(void* const& buffer, const ulong& stride, const real_t& p, const real_t& q, const ulong outerSegments, const ulong& innerSegments) throw();

//! Utility function to create mipmaps for a normal map.
void FilterNormalMap(const Texture& texture) throw(Exception);

//! Utility function to create an attenuation volume texture.
void ComputeAttenuationVolume(const VolumeTexture& volume) throw(Exception);
//! Utility function to create an attenuation volume texture.
void ComputeAttenuationVolumeIndirect(const Graphics& graphics, const VolumeTexture& volume) throw(Exception);

//! Utility function to evaluate the texels of a cube texture face using the given functor.
template<class Evaluator>
void ComputeCubeTextureFace(const D3DLOCKED_RECT& lock, const ulong& faceLength, const Vector3& base, const Vector3& uOffset, const Vector3& vOffset, Evaluator& eval) throw()
{
	for(ulong y = 0; y < faceLength; ++y)
	{
		ulong* buffer = reinterpret_cast<ulong*>(reinterpret_cast<size_t>(lock.pBits) + y * lock.Pitch);
		for(ulong x = 0; x < faceLength; ++x)
		{
            const Vector3 ray = base + ((_R(x) + _R(0.5)) * uOffset + (_R(y) + _R(0.5)) * vOffset) / _R(faceLength);
			buffer[x] = eval(ray);
		}
	}
}

//! Utility function to evaluate the texels of a cube texture using the given functor.
template<class Evaluator>
void ComputeCubeTexture(const CubeTexture& cube, const Evaluator& eval) throw(Exception)
{
	// get the cube side length
    D3DSURFACE_DESC desc;
	HRESULT hr = cube->GetLevelDesc(0, &desc);
	if(FAILED(hr))
		throw CreateDirectXException("ComputeCubeTexture(): GetLevelDesc failed: ", hr);

	// lock and compute each face
	D3DLOCKED_RECT lock;
	hr = cube->LockRect(D3DCUBEMAP_FACE_POSITIVE_X, 0, &lock, 0, 0);
	if(FAILED(hr))
		throw CreateDirectXException("ComputeCubeTexture(): LockRect failed: ", hr);
	ComputeCubeTextureFace(lock, desc.Width, 
		Vector3(_R(1), _R(1), _R(1)), 
		Vector3(_R(0), _R(0), _R(-2)),   
		Vector3(_R(0), _R(-2), _R(0)), 
		eval
	);
	cube->UnlockRect(D3DCUBEMAP_FACE_POSITIVE_X, 0);

	hr = cube->LockRect(D3DCUBEMAP_FACE_NEGATIVE_X, 0, &lock, 0, 0);
	if(FAILED(hr))
		throw CreateDirectXException("ComputeCubeTexture(): LockRect failed: ", hr);
	ComputeCubeTextureFace(lock, desc.Width, 
		Vector3(_R(-1), _R(1), _R(-1)), 
		Vector3(_R(0), _R(0), _R(2)),   
		Vector3(_R(0), _R(-2), _R(0)), 
		eval
	);
	cube->UnlockRect(D3DCUBEMAP_FACE_NEGATIVE_X, 0);

	hr = cube->LockRect(D3DCUBEMAP_FACE_POSITIVE_Y, 0, &lock, 0, 0);
	if(FAILED(hr))
		throw CreateDirectXException("ComputeCubeTexture(): LockRect failed: ", hr);
	ComputeCubeTextureFace(lock, desc.Width, 
		Vector3(_R(-1), _R(1), _R(-1)), 
		Vector3(_R(2), _R(0), _R(0)),   
		Vector3(_R(0), _R(0), _R(2)), 
		eval
	);
	cube->UnlockRect(D3DCUBEMAP_FACE_POSITIVE_Y, 0);

	hr = cube->LockRect(D3DCUBEMAP_FACE_NEGATIVE_Y, 0, &lock, 0, 0);
	if(FAILED(hr))
		throw CreateDirectXException("ComputeCubeTexture(): LockRect failed: ", hr);
	ComputeCubeTextureFace(lock, desc.Width, 
		Vector3(_R(-1), _R(-1), _R(1)), 
		Vector3(_R(2), _R(0), _R(0)),   
		Vector3(_R(0), _R(0), _R(-2)), 
		eval
	);
	cube->UnlockRect(D3DCUBEMAP_FACE_NEGATIVE_Y, 0);

	hr = cube->LockRect(D3DCUBEMAP_FACE_POSITIVE_Z, 0, &lock, 0, 0);
	if(FAILED(hr))
		throw CreateDirectXException("ComputeCubeTexture(): LockRect failed: ", hr);
	ComputeCubeTextureFace(lock, desc.Width, 
		Vector3(_R(-1), _R(1), _R(1)), 
		Vector3(_R(2), _R(0), _R(0)),   
		Vector3(_R(0), _R(-2), _R(0)), 
		eval
	);
	cube->UnlockRect(D3DCUBEMAP_FACE_POSITIVE_Z, 0);

	hr = cube->LockRect(D3DCUBEMAP_FACE_NEGATIVE_Z, 0, &lock, 0, 0);
	if(FAILED(hr))
		throw CreateDirectXException("ComputeCubeTexture(): LockRect failed: ", hr);
	ComputeCubeTextureFace(lock, desc.Width, 
		Vector3(_R(1), _R(1), _R(-1)), 
		Vector3(_R(-2), _R(0), _R(0)),   
		Vector3(_R(0), _R(-2), _R(0)), 
		eval
	);
	cube->UnlockRect(D3DCUBEMAP_FACE_NEGATIVE_Z, 0);
}

//! Utility function to indirectly evaluate the texels of a cube texture face using the given functor.
template<class Evaluator>
void ComputeCubeTextureIndirect(const Graphics& graphics, const CubeTexture& cube, const Evaluator& eval) throw(Exception)
{
	// get the cube side length
    D3DSURFACE_DESC desc;
	HRESULT hr = cube->GetLevelDesc(0, &desc);
	if(FAILED(hr))
		throw CreateDirectXException("ComputeCubeTextureIndirect(): GetLevelDesc failed: ", hr);

	// create a system memory version
	CubeTexture temp;
	hr = graphics->CreateCubeTexture(desc.Width, 1, 0, desc.Format, D3DPOOL_SYSTEMMEM, temp.AddressOf());
	if(FAILED(hr))
		throw CreateDirectXException("ComputeCubeTextureIndirect(): CreateCubeTexture failed: ", hr);

	// compute the cube texture
	ComputeCubeTexture(temp, eval);

	// copy to the hardware one
	for(ulong i = 0; i < 6; ++i)
	{
		hr = cube->AddDirtyRect(static_cast<D3DCUBEMAP_FACES>(i), 0);
		if(FAILED(hr))
			throw CreateDirectXException("ComputeCubeTextureIndirect(): AddDirtyRect failed: ", hr);
	}
	hr = graphics->UpdateTexture(temp.Get(), cube.Get());
	if(FAILED(hr))
		throw CreateDirectXException("ComputeCubeTextureIndirect(): UpdateTexture failed: ", hr);
}

//! Utility evaluation functor that returns the given ray normalised as an RGB colour.
class NormalisationEvaluator
{
public:
	//! Evaluates a ray through a texel, returning the normalised value as an RGB colour.
	ulong operator()(const Vector3& ray) const throw();
};

//! Utility evaluator functor that evaluates a ray using the specular response intensity.
class SpecularEvaluator
{
private:
	real_t power;

public:
	//! Creates a new specular evaluator.
	SpecularEvaluator(const real_t& power_) throw();

	//! Evaluates a ray through a texel, returning the specular response.
	ulong operator()(const Vector3& ray) const throw();
};

//! Utility function to create a copy of a given system memory texture in the default memory pool.
Texture CopyTextureToDefaultPool(const Graphics& graphics, const Texture& texture) throw(Exception);

} // namespace oe