/*  OpenEngine: A free open source rendering system for DirectX.
	Copyright (C) 2002 Simon Brown

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <oe/openengine.h>

class Cube
{
private:
	oe::OrientedVertexBuffer vertices;
	oe::TexCoordBuffer texcoords;

public:
	explicit Cube(oe::Graphics& graphics) throw(oe::Exception)
	{
		// allocate resources
		vertices = graphics.CreateVertexBuffer<oe::OrientedVertex>(36, oe::OEBT_STATIC);
		texcoords = graphics.CreateVertexBuffer<oe::TexCoord>(36, oe::OEBT_STATIC);
		// compute the vertices
		oe::OrientedVertexBuffer::SafeLock lock = vertices.Lock();
		oe::CreateCubeTriangleListVertices(&lock[0].p, sizeof(oe::OrientedVertex));
		oe::CreateCubeTriangleListNormals(&lock[0].n, sizeof(oe::OrientedVertex));
		oe::CreateCubeTriangleListTangents(&lock[0].t, sizeof(oe::OrientedVertex));
		// compute the texcoords
		oe::TexCoordBuffer::SafeLock texlock = texcoords.Lock();
		oe::CreateCubeTriangleListTexcoords(texlock.Get(), sizeof(oe::TexCoord));
	}

	oe::OrientedVertexBuffer GetVertices() const
	{
		return vertices;
	}
	oe::TexCoordBuffer GetTexCoords() const
	{
		return texcoords;
	}
};

class CubeRef : public oe::Renderable
{
private:
	oe::Vector3 centre, size;

public:
	CubeRef(const Cube& cube_, 
		    const oe::Material& material_, 
			const oe::Vector3& centre_, 
			const oe::Vector3& size_) throw() 
	  : Renderable(oe::AABB(centre_ - _R(0.5) * size_, centre_ + _R(0.5) * size_), 
				   material_, 
				   cube_.GetVertices(), 
				   cube_.GetTexCoords()
		),
		centre(centre_), 
		size(size_)
	{
		// create the local transform
		CreateTransform();
	}

	void SetOrientation(const oe::Quaternion& orientation_) throw()
	{
		orientation = orientation_;
	}

	void CreateTransform() throw()
	{
		// compute the local transform
		transform.CreateFromScale(_R(0.5) * size);
		transform = transform * oe::Matrix4x4(orientation);
		transform.PostTranslate(centre);
	}

	virtual void DrawPrimitive(const oe::Graphics& graphics) const throw(oe::Exception) 
	{ 
		graphics->DrawPrimitive(D3DPT_TRIANGLELIST, 0, 12); 
	}
};

void RenderText(const oe::Graphics& graphics, const oe::VertexShader& guiVS, const oe::PixelShader& guiPS, const oe::TextBuffer& text)
{
	graphics->SetRenderState(D3DRS_ZFUNC, D3DCMP_ALWAYS);
	graphics->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
	graphics->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
	graphics->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
	graphics->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);

	graphics.SetVertexShader(guiVS);
	graphics.SetPixelShader(guiPS);

	oe::Matrix4x4 guiTransform;
	oe::CreateGuiTransform(graphics, guiTransform);
	static const float textColour[] = { 1.0f, 1.0f, 1.0f, 1.0f };

	graphics->SetVertexShaderConstant(VS_PROJECTION, &guiTransform, 4);
	graphics->SetPixelShaderConstant(PS_LIGHTCOLOUR, textColour, 1);

    text.DrawPrimitive(graphics);

	graphics->SetRenderState(D3DRS_ZFUNC, D3DCMP_LESSEQUAL);
	graphics->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);
	graphics->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
}

void oe_main(oe::Graphics& graphics) throw(oe::Exception)
{
	// show the loading screen and create the device
	graphics.SetDepthStencilFormat(D3DFMT_D16);
	if(!graphics.CreateFromLoader("OpenEngine Demo"))
		return;

	// load up a directinput device
	oe::Input input(graphics);

	// create the gui objects
	oe::VertexShader guiVS = oe::CreateGuiVertexShader(graphics);
	oe::PixelShader guiPS = oe::CreateGuiPixelShader(graphics);
	oe::Font guiFont = graphics.LoadFont("fonts/gui");
	oe::TextBuffer stats(graphics, 400);
	oe::TextBuffer controls(graphics, 400);
	stats.SetFont(guiFont);
	controls.SetFont(guiFont);

	controls.SetPosition(10, 10);
	controls.Println("rotate camera: up/down/left/right");
	controls.Println("rotate light instead: hold L");
	controls.Println("wireframe: hold W");

	// create our sample scene
	oe::Scene scene(graphics);

	// create a material
	oe::Material material;
	oe::Texture temp = graphics.LoadTexture("textures/metal_normal.tga", D3DFMT_X8R8G8B8, D3DPOOL_SYSTEMMEM, true);
	oe::FilterNormalMap(temp);
	material.normals = oe::CopyTextureToDefaultPool(graphics, temp);
	material.diffuse = graphics.LoadTexture("textures/metal_diffuse.tga", D3DFMT_X8R8G8B8, D3DPOOL_DEFAULT, true);
	material.diffuseIntensity = _R(0.7);
	HRESULT hr = graphics->CreateCubeTexture(128, 1, 0, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, material.specular.AddressOf());
	if(FAILED(hr))
		throw oe::CreateDirectXException("CreateCubeTexture() failed: ", hr);
	oe::ComputeCubeTextureIndirect(graphics, material.specular, oe::SpecularEvaluator(_R(20)));
	material.specularIntensity = _R(0.7);

	// create a grid of cubes with this material
	Cube cube(graphics);
	for(oe::ulong z = 0; z < 3; ++z)
	{
		for(oe::ulong x = 0; x < 3; ++x)
		{
			boost::shared_ptr<CubeRef> cubeRef(new CubeRef(
				cube, 
				material, 
				oe::Vector3(_R(2 * x), _R(0), _R(2 * z)) - oe::Vector3(_R(2), _R(0), _R(2)), 
				oe::Vector3(_R(1.5), _R(1.5), _R(1.5))
			));
			scene.AddRenderable(cubeRef);
		}
	}

	// create a spotlight
	boost::shared_ptr<oe::SpotLight> spot(new oe::SpotLight);
	spot->colour = oe::Vector3(_R(1), _R(1), _R(1));
	spot->intensity = graphics.LoadTexture("textures/spot_intensity.tga", D3DFMT_A8, D3DPOOL_DEFAULT, true);
	spot->fovRatio = _R(2.5);
	spot->aspectRatio = _R(1);
	spot->range = _R(15);
	static const oe::ulong shadowmapsize = 512;
	hr = graphics->CreateTexture(shadowmapsize, shadowmapsize, 1, D3DUSAGE_RENDERTARGET, D3DFMT_X8R8G8B8, D3DPOOL_DEFAULT, spot->rendertarget.AddressOf());
	if(FAILED(hr))
		throw oe::CreateDirectXException("CreateTexture() failed: ", hr);
	hr = graphics->CreateTexture(shadowmapsize, shadowmapsize, 1, D3DUSAGE_DEPTHSTENCIL, D3DFMT_D24X8, D3DPOOL_DEFAULT, spot->shadowmap.AddressOf());
	if(FAILED(hr))
		throw oe::CreateDirectXException("CreateTexture() failed: ", hr);
	spot->CreateTransforms();
	scene.AddSpotLight(spot);

	// create the target projection
	oe::Projection projection;
	projection.aspectRatio = graphics.GetTargetAspectRatio();
	projection.CreateTransform();

	oe::Quaternion viewOrientation(oe::Vector3(_R(1), _R(0), _R(0)), -oe::RealPi / _R(6));
	oe::Quaternion lightOrientation(viewOrientation);
	oe::Vector3 camera(_R(0), _R(0), _R(-9));

	oe::real_t frameTimer = _R(0), fps = _R(0);
	oe::ulong frameCounter = 0;

	// the main application loop
	for(;;)
	{
		if(graphics.ProcessThreadMessages())
			break;

		oe::real_t frameTime = graphics.GetFrameTimer().Mark();
		frameTimer += frameTime;
		++frameCounter;
		if(frameTimer > _R(0.25))
		{
			fps = _R(frameCounter) / frameTimer;
			frameTimer = _R(0);
			frameCounter = 0;
		}

		input.GetImmediateKeyboardData();

		// use view or light orientation based on the input
		oe::Quaternion&	povOrientation = input.IsKeyDown(DIK_L) ? lightOrientation : viewOrientation;
		
		// rotate the light based on the keyboard input
		const oe::real_t amt = frameTime * oe::RealPi / _R(2);
		povOrientation = oe::Quaternion(oe::Vector3(_R(1), _R(0), _R(0)), 
			((input.IsKeyDown(DIK_UP) ? _R(-1) : _R(0)) + (input.IsKeyDown(DIK_DOWN) ? _R(1) : _R(0))) * amt
		) * povOrientation;
		povOrientation = povOrientation * oe::Quaternion(oe::Vector3(_R(0), _R(1), _R(0)), 
			((input.IsKeyDown(DIK_RIGHT) ? _R(1) : _R(0)) + (input.IsKeyDown(DIK_LEFT) ? _R(-1) : _R(0))) * amt
		);
		povOrientation /= povOrientation.GetNorm();

		// compute the pov position
		oe::Vector3 povPosition = oe::UnitQuaternionRotation(~povOrientation, camera);

		// update the spotlight
		spot->orientation = lightOrientation;
		spot->position = oe::UnitQuaternionRotation(~lightOrientation, camera);
		spot->CreateTransforms();

		// render the shadow maps
		scene.RenderShadowMaps(graphics);

		// set the view and batch up the scene
		scene.SetView(povPosition, povOrientation, projection);
		scene.Compile();

		// render the scene
		graphics->Clear(0, 0, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0, 1.0f, 0);
		graphics->BeginScene();

		graphics->SetRenderState(D3DRS_FILLMODE, input.IsKeyDown(DIK_W) ? D3DFILL_WIREFRAME : D3DFILL_SOLID);
		scene.Render(graphics);
		
		graphics->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
		RenderText(graphics, guiVS, guiPS, controls);
		RenderText(graphics, guiVS, guiPS, stats);

		graphics->EndScene();
		graphics->Present(0, 0, 0, 0);

		// update the stats
		stats.Clear();
		stats.SetPosition(static_cast<short>(graphics.GetTargetWidth() - 150), 10);
		std::ostringstream ss;
		ss << std::setiosflags(std::ios::fixed) << std::setprecision(1) << fps << " fps";
		stats.Println(ss.str());
		stats.Println(input.IsKeyDown(DIK_L) ? "light POV" : "camera POV");
		stats.Println();
		ss.str("");
		ss << "Fragments: " << scene.GetNumEntityPasses();
		stats.Println(ss.str());
		ss.str("");
		ss << "State changes: " << scene.GetNumStateChanges();
		stats.Println(ss.str());
	}
}